"""Contains functions that are used in the dynamic location and creation 
of tabs and datatypes. Can be used both internally and externally.
"""

import os
import json
import pkg_resources
import importlib
import logging

from PyQt5 import QtWidgets

from meggie.utilities.uid import generate_uid
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.threading import threaded


def find_all_plugins():
    """Looks for plugins (installed packages with name meggie_*), that
    can contain tabs or datatypes"""
    plugins = []
    package_keys = [dist.key.replace('-', '_') for dist 
                    in pkg_resources.working_set]
    for key in package_keys:
        try:
            if key.startswith('meggie_'):
                # check that there exists configuration.json
                if not os.path.exists(
                        pkg_resources.resource_filename(key, 'configuration.json')):
                    continue
                plugins.append(key)
        except Exception as exc:
            logging.getLogger('ui_logger').exception('')
    return plugins


def find_all_sources():
    """Returns all packages where to look for actions / datatypes.
    """
    return ['meggie'] + find_all_plugins()


def find_all_package_specs():
    """Returns all package specifications found.
    """
    package_specs = {}

    sources = find_all_sources()
    for source in sources:
        config_path = pkg_resources.resource_filename(source, 'configuration.json')
        if not os.path.exists(config_path):
            continue
        with open(config_path, 'r') as f:
            config = json.load(f)
            if config:
                package_specs[source] = config

    # add possibly missing fields
    for package_spec in package_specs.values():
        if 'name' not in package_spec:
            package_spec['name'] = ""

        if 'author' not in package_spec:
            package_spec['author'] = ""

    return package_specs


def find_all_datatype_specs():
    """Returns all datatype specifications found.
    """
    datatype_specs = {}
    found_keys = []

    sources = find_all_sources()
    for source in sources:
        datatype_path = pkg_resources.resource_filename(source, 'datatypes')
        if not os.path.exists(datatype_path):
            continue
        for package in os.listdir(datatype_path):
            config_path = os.path.join(
                datatype_path, package, 'configuration.json')
            if os.path.exists(config_path):
                with open(config_path, 'r') as f:
                    config = json.load(f)
                    if config:
                        if config['id'] in found_keys:
                            raise Exception('Datatype with the same id found ' + 
                                            'in multiple packages')
                        else:
                            found_keys.append(config['id'])

                        datatype_specs[config['id']] = source, package, config

    # add possibly missing fields
    for source, package, datatype_spec in datatype_specs.values():
        if 'name' not in datatype_spec:
            datatype_spec['name'] = datatype_spec['id']
        
    return datatype_specs


def find_all_action_specs():
    """Returns all action specifications found.
    """
    action_specs = {}
    found_keys = []

    sources = find_all_sources()
    for source in sources:
        action_path = pkg_resources.resource_filename(source, 'actions')
        if not os.path.exists(action_path):
            continue
        for package in os.listdir(action_path):
            config_path = os.path.join(action_path, package, 'configuration.json')
            if os.path.exists(config_path):
                with open(config_path, 'r') as f:
                    config = json.load(f)
                    if config:
                        if config['id'] in found_keys:
                            raise Exception('Action with the same id found ' + 
                                            'in multiple packages')
                        else:
                            found_keys.append(config['id'])

                        action_specs[config['id']] = source, package, config

    # add possibly missing fields
    for source, package, action_spec in action_specs.values():

        if 'name' not in action_spec:
            action_spec['name'] = action_spec['id']
        
    return action_specs


def construct_tab(tab_spec, action_specs, datatype_specs, parent):
    """Constructs analysis tab dynamically. Returns a QDialog
    that can be used within a QTabDialog of the main window.

    Parameters
    ----------
    tab_spec : dict
        The specification of the tab read to a dict
    action_specs : dict
        Specifications of actions
    datatype_specs : dict
        Specifications of datatypes
    parent : instance of main window
        The main window, is passed to the handlers in the ui.py.

    Returns
    -------
    instance of QDialog
        The constructed tab than can be added to
        main window.
    """

    class DynamicTab(QtWidgets.QDialog):
        """ Class defining a tab.
        
        Parameters
        ----------
        parent : instance of main window
            The main window, is passed to the handlers in actions.
        """
        def __init__(self, parent):
            QtWidgets.QDialog.__init__(self)
            self.parent = parent
            self.tab_spec = tab_spec
            self.action_specs = action_specs
            self.datatype_specs = datatype_specs

            # first create basic layout
            self.gridLayoutContainer = QtWidgets.QGridLayout(self)

            self.gridLayoutRoot = QtWidgets.QGridLayout()
            self.gridLayoutContainer.addLayout(self.gridLayoutRoot, 0, 0, 1, 1)

            if self.tab_spec['outputs']:
                self.groupBoxOutputs = QtWidgets.QGroupBox(self)
                self.groupBoxOutputs.setTitle('Outputs')
                self.gridLayoutOutputs = QtWidgets.QGridLayout(
                    self.groupBoxOutputs)

            if self.tab_spec['output_actions']:
                self.groupBoxOutputActions = QtWidgets.QGroupBox(self)
                self.groupBoxOutputActions.setTitle('')
                self.gridLayoutOutputActions = QtWidgets.QGridLayout(
                    self.groupBoxOutputActions)

            if self.tab_spec['input_actions']:
                self.groupBoxInputActions = QtWidgets.QGroupBox(self)
                self.groupBoxInputActions.setTitle('')
                self.gridLayoutInputActions = QtWidgets.QGridLayout(
                    self.groupBoxInputActions)

            if self.tab_spec['inputs']:
                self.groupBoxInputs = QtWidgets.QGroupBox(self)
                self.groupBoxInputs.setTitle('Inputs')
                self.gridLayoutInputs = QtWidgets.QGridLayout(
                    self.groupBoxInputs)

            if self.tab_spec['info']:
                self.groupBoxInfo = QtWidgets.QGroupBox(self)
                self.groupBoxInfo.setTitle('Info')
                self.gridLayoutInfo = QtWidgets.QGridLayout(self.groupBoxInfo)

            # add the (empty) input lists
            for idx, input_name in enumerate(self.tab_spec['inputs']):

                title = input_name.capitalize()

                for source, package, datatype_spec in datatype_specs.values():
                    if datatype_spec['id'] == input_name:
                        title = datatype_spec['name']

                groupBoxInputElement = QtWidgets.QGroupBox(self.groupBoxInputs)
                groupBoxInputElement.setTitle(title)
                gridLayoutInputElement = QtWidgets.QGridLayout(
                    groupBoxInputElement)
                listWidgetInputElement = QtWidgets.QListWidget(
                    groupBoxInputElement)

                listWidgetInputElement.setSelectionMode(
                    QtWidgets.QAbstractItemView.ExtendedSelection)

                gridLayoutInputElement.addWidget(
                    listWidgetInputElement, idx, 0, 1, 1)
                self.gridLayoutInputs.addWidget(groupBoxInputElement)

                setattr(self, 'groupBoxInputElement_' + str(idx + 1),
                        groupBoxInputElement)
                setattr(self, 'gridLayoutInputElement_' + str(idx + 1),
                        gridLayoutInputElement)
                setattr(self, 'listWidgetInputElement_' + str(idx + 1),
                        listWidgetInputElement)

            # add the (empty) output lists
            for idx, output_name in enumerate(self.tab_spec['outputs']):

                title = output_name.capitalize()

                for source, package, datatype_spec in datatype_specs.values():
                    if datatype_spec['id'] == output_name:
                        title = datatype_spec['name']

                groupBoxOutputElement = QtWidgets.QGroupBox(
                    self.groupBoxOutputs)
                groupBoxOutputElement.setTitle(title)
                gridLayoutOutputElement = QtWidgets.QGridLayout(
                    groupBoxOutputElement)
                listWidgetOutputElement = QtWidgets.QListWidget(
                    groupBoxOutputElement)

                listWidgetOutputElement.setSelectionMode(
                    QtWidgets.QAbstractItemView.ExtendedSelection)

                gridLayoutOutputElement.addWidget(
                    listWidgetOutputElement, idx, 0, 1, 1)
                self.gridLayoutOutputs.addWidget(groupBoxOutputElement)

                setattr(self, 'groupBoxOutputElement_' + str(idx + 1),
                        groupBoxOutputElement)
                setattr(self, 'gridLayoutOutputElement_' + str(idx + 1),
                        gridLayoutOutputElement)
                setattr(self, 'listWidgetOutputElement_' + str(idx + 1),
                        listWidgetOutputElement)

            # add input action buttons
            for idx, action_name in enumerate(self.tab_spec['input_actions']):

                action_spec = self.action_specs[action_name][2]
                title = action_spec['name']
                
                pushButtonInputActionElement = QtWidgets.QPushButton(
                    self.groupBoxInputActions)
                pushButtonInputActionElement.setText(title)
                self.gridLayoutInputActions.addWidget(
                    pushButtonInputActionElement, idx, 0, 1, 1)
                setattr(self, 'pushButtonInputActionElement_' + str(idx + 1),
                        pushButtonInputActionElement)

            if getattr(self, 'gridLayoutInputActions', None):
                spacer = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum,
                                               QtWidgets.QSizePolicy.Expanding)
                self.gridLayoutInputActions.addItem(spacer, idx + 1, 0, 1, 1)

            # add action buttons
            for idx, action_name in enumerate(self.tab_spec['output_actions']):

                action_spec = self.action_specs[action_name][2]
                title = action_spec['name']
 
                pushButtonOutputActionElement = QtWidgets.QPushButton(
                    self.groupBoxOutputActions)
                pushButtonOutputActionElement.setText(title)
                self.gridLayoutOutputActions.addWidget(
                    pushButtonOutputActionElement, idx, 0, 1, 1)
                setattr(self, 'pushButtonOutputActionElement_' + str(idx + 1),
                        pushButtonOutputActionElement)

            if getattr(self, 'gridLayoutOutputActions', None):
                spacer = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum,
                                               QtWidgets.QSizePolicy.Expanding)
                self.gridLayoutOutputActions.addItem(spacer, idx + 1, 0, 1, 1)

            # add info text boxes
            for idx, info_name in enumerate(self.tab_spec['info']):

                action_spec = self.action_specs[info_name][2]
                title = action_spec['name']

                groupBoxInfoElement = QtWidgets.QGroupBox(self.groupBoxInfo)
                groupBoxInfoElement.setTitle(title)
                gridLayoutInfoElement = QtWidgets.QGridLayout(
                    groupBoxInfoElement)
                plainTextEditInfoElement = QtWidgets.QTextBrowser(
                    groupBoxInfoElement)

                gridLayoutInfoElement.addWidget(
                    plainTextEditInfoElement, idx, 0, 1, 1)
                self.gridLayoutInfo.addWidget(groupBoxInfoElement)

                setattr(self, 'groupBoxInfoElement_' + str(idx + 1),
                        groupBoxInfoElement)
                setattr(self, 'gridLayoutInfoElement_' + str(idx + 1),
                        gridLayoutInfoElement)
                setattr(self, 'plainTextEditInfoElement_' + str(idx + 1),
                        plainTextEditInfoElement)

            # lay out inputs, outputs, actions and info elements
            # in a nice way to a grid
            if self.tab_spec['inputs'] and not self.tab_spec['input_actions']:
                self.gridLayoutRoot.addWidget(self.groupBoxInputs, 0, 0, 2, 1)
            elif self.tab_spec['inputs']:
                self.gridLayoutRoot.addWidget(self.groupBoxInputs, 0, 0, 1, 1)

            if self.tab_spec['outputs'] and not self.tab_spec['output_actions']:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputs, 0, 1, 2, 1)
            elif self.tab_spec['outputs']:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputs, 0, 1, 1, 1)

            if self.tab_spec['input_actions'] and not self.tab_spec['inputs']:
                self.gridLayoutRoot.addWidget(
                    self.groupBoxInputActions, 0, 0, 2, 1)
            elif self.tab_spec['input_actions']:
                self.gridLayoutRoot.addWidget(
                    self.groupBoxInputActions, 1, 0, 1, 1)

            if self.tab_spec['output_actions'] and not self.tab_spec['outputs']:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputActions, 0, 1, 2, 1)
            elif self.tab_spec['output_actions']:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputActions, 1, 1, 1, 1)

            if self.tab_spec['info']:
                self.gridLayoutRoot.addWidget(self.groupBoxInfo, 0, 2, 2, 1)

            # add spacers to bottom and right to keep the window concise
            spacerItemVertical = QtWidgets.QSpacerItem(
                20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
            self.gridLayoutContainer.addItem(spacerItemVertical, 1, 0, 1, 1)
            spacerItemHorizontal = QtWidgets.QSpacerItem(
                10, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
            self.gridLayoutContainer.addItem(spacerItemHorizontal, 0, 1, 1, 1)

            # add handlers for list selection changed -> info updates
            def connect_to_handler(list_element, info_element, info_name):

                source, package, action_spec = self.action_specs[info_name]
                module = importlib.import_module(
                    '.'.join([source, 'actions', package]))
                entry = action_spec['entry']

                @threaded
                def handler(*args):
                    return getattr(module, entry)(*args).run()

                def handler_wrapper():
                    experiment = self.parent.experiment
                    if not experiment:
                        return
                    subject = experiment.active_subject
                    if not subject:
                        return

                    data = self._get_data()

                    try:
                        info_content = handler(experiment, data, 
                                               parent, action_spec, 
                                               do_meanwhile=parent.update_ui)
                        info_element.setPlainText(info_content)
                    except Exception as exc:
                        exc_messagebox(self, exc)

                list_element.itemSelectionChanged.connect(handler_wrapper)

            for idx, info_name in enumerate(self.tab_spec['info']):
                info_element = getattr(self, 'plainTextEditInfoElement_' +
                                       str(idx + 1))

                for idx, input_name in enumerate(self.tab_spec['inputs']):

                    input_element = getattr(
                        self, 'listWidgetInputElement_' + str(idx + 1))
                    connect_to_handler(input_element, info_element, info_name)

                for idx, output_name in enumerate(self.tab_spec['outputs']):

                    output_element = getattr(
                        self, 'listWidgetOutputElement_' + str(idx + 1))
                    connect_to_handler(output_element, info_element, info_name)

            # add button handlers
            def connect_to_handler(button, name):

                source, package, action_spec = self.action_specs[name]
                module = importlib.import_module(
                    '.'.join([source, 'actions', package]))
                entry = action_spec['entry']

                handler = getattr(module, entry)

                def handler_wrapper(checked):
                    experiment = self.parent.experiment
                    if not experiment:
                        return

                    subject = experiment.active_subject
                    if not subject:
                        return

                    data = self._get_data()

                    try:
                        handler(experiment, data, parent, action_spec).run()
                    except Exception as exc:
                        exc_messagebox(self, exc)

                button.clicked.connect(handler_wrapper)

            for idx, action_name in enumerate(self.tab_spec['input_actions']):
                action_element = getattr(
                    self, 'pushButtonInputActionElement_' + str(idx + 1))
                connect_to_handler(action_element, action_name)

            for idx, action_name in enumerate(self.tab_spec['output_actions']):
                action_element = getattr(
                    self, 'pushButtonOutputActionElement_' + str(idx + 1))
                connect_to_handler(action_element, action_name)
 

        def _get_data(self):
            """Returns data from input and output lists.
            """
            data = {'inputs': {},
                    'outputs': {}}

            data['tab_id'] = self.tab_spec['id']

            inputs = []
            for idx, name in enumerate(self.tab_spec['inputs']):
                ui_element = getattr(
                    self, 'listWidgetInputElement_' + str(idx + 1))
                try:
                    selected_items = [item.text()
                                      for item in ui_element.selectedItems()]
                except BaseException:
                    continue

                data['inputs'][name] = selected_items

            for idx, name in enumerate(self.tab_spec['outputs']):
                ui_element = getattr(
                    self, 'listWidgetOutputElement_' + str(idx + 1))
                try:
                    selected_items = [item.text() for item in
                                      ui_element.selectedItems()]
                except BaseException:
                    continue

                data['outputs'][name] = selected_items

            return data

        def initialize_ui(self):
            """Updates (empties and refills) ui contents when called.
            """
            experiment = self.parent.experiment
            if not experiment:
                return

            subject = experiment.active_subject

            # fill input lists with contents
            for idx, input_name in enumerate(self.tab_spec['inputs']):
                ui_element = getattr(
                    self, 'listWidgetInputElement_' + str(idx + 1))

                ui_element.clear()

                if not subject:
                    continue

                data = getattr(subject, input_name, None)
                if not data:
                    continue

                for key in sorted(list(data.keys())):
                    ui_element.addItem(key)

            # fill output lists with contents
            for idx, output_name in enumerate(
                    self.tab_spec['outputs']):
                ui_element = getattr(
                    self, 'listWidgetOutputElement_' + str(idx + 1))

                ui_element.clear()

                if not subject:
                    continue

                data = getattr(subject, output_name, None)
                if not data:
                    continue

                for key in sorted(list(data.keys())):
                    ui_element.addItem(key)

            # allow to fill info element already here.
            # there are also handlers to update info element
            # on list selection changes
            for idx, info_name in enumerate(self.tab_spec['info']):

                ui_element = getattr(
                    self, 'plainTextEditInfoElement_' + str(idx + 1))

                source, package, action_spec = self.action_specs[info_name]
                module = importlib.import_module(
                    '.'.join([source, 'actions', package]))
                entry = action_spec['entry']

                @threaded
                def handler(*args):
                    return getattr(module, entry)(*args).run()

                try:
                    info_content = handler(experiment, None, 
                                           self.parent, action_spec, 
                                           do_meanwhile=parent.update_ui)
                    ui_element.setPlainText(info_content)
                except Exception as exc:
                    exc_messagebox(self, exc)


        @property
        def name(self):
            """Returns name of the tab."""
            return self.tab_spec['name']

    DynamicTab.__name__ = 'MainWindowTab' + tab_spec['id'].capitalize()

    tab = DynamicTab(parent)

    return tab


def construct_tabs(selected_pipeline, window, prefs, include_eeg):
    """Constructs as set of tabs based on specifications and the
    selected pipeline.

    Parameters
    ----------
    selected_pipeline : str
        ID of the selected pipeline
    window : instance of main window
        The main window.
    prefs : Instance of PreferencesHandler
        Stores e.g. active plugins.
    include_eeg : bool
        Whether to add EEG-related actions

    Returns
    -------
    list of QDialog
        Contains the constructed tabs relevant to the pipeline
    """
 
    active_plugins = prefs.active_plugins

    action_specs = find_all_action_specs()
    datatype_specs = find_all_datatype_specs()
    package_specs = find_all_package_specs()

    tabs = []
    pipelines = []
    for source, package_spec in package_specs.items():
        if source not in active_plugins and source != "meggie":
            continue
        if 'tabs' in package_spec:
            tabs.extend(package_spec['tabs'])
        if 'pipelines' in package_spec:
            pipelines.extend(package_spec['pipelines'])

    # add possibly missing fields
    for tab in tabs:
        if 'id' not in tab:
            raise Exception('Every tab specification must have id.')

        if 'name' not in tab:
            tab['name'] = tab['id']

        if 'inputs' not in tab:
            tab['inputs'] = []

        if 'outputs' not in tab:
            tab['outputs'] = []

        if 'input_actions' not in tab:
            tab['input_actions'] = []

        if 'output_actions' not in tab:
            tab['output_actions'] = []

        if 'info' not in tab:
            tab['info'] = []
 
    for pipeline in pipelines:
        if 'id' not in pipeline:
            raise Exception('Every pipeline must have id.')

        if 'name' not in pipeline:
            pipeline['name'] = pipeline['id']

    found = False
    pipeline_spec = None
    for pipeline in pipelines:
        if pipeline['id'] == selected_pipeline:
            found = True
            pipeline_spec = pipeline
            break
    if not found:
        # Use classic
        pipeline_spec = {'id': "classic",
                         'name': "Include everything"}

    # merges tab specification from others to first and 
    # filters to tabs specified by the pipeline
    combined_tabs = []
    for tab_spec in tabs:
        # Include only tabs relevant to the pipeline
        if pipeline_spec.get("include_tabs"):
            if tab_spec['id'] not in pipeline_spec['include_tabs']:
                continue

        # if a completely new tab, initialize it
        if tab_spec['id'] not in [tab['id'] for tab in combined_tabs]:

            new_tab = {}
            new_tab['id'] = tab_spec['id']
            new_tab['name'] = tab_spec['name']
            new_tab['inputs'] = []
            new_tab['outputs'] = []
            new_tab['input_actions'] = []
            new_tab['output_actions'] = []
            new_tab['info'] = []

            idx = len(combined_tabs)
            combined_tabs.append(new_tab)
        # otherwise find idx of the tab in the first specification
        else:
            idx = [tab['id'] for tab in combined_tabs].index(tab_spec['id'])

        # Add missing inputs
        for input_spec in tab_spec['inputs']:
            if input_spec not in combined_tabs[idx]['inputs']:
                combined_tabs[idx]['inputs'].append(input_spec)

        # Add missing outputs
        for output_spec in tab_spec['outputs']:
            if output_spec not in combined_tabs[idx]['outputs']:
                combined_tabs[idx]['outputs'].append(output_spec)

        # add missing input actions
        for input_spec in tab_spec['input_actions']:

            action_spec = action_specs.get(input_spec)
            if not action_spec:
                raise Exception("Cannot read action " + input_spec + ".")

            if not include_eeg and 'eeg' in action_spec[2].get('tags', []):
                continue

            if input_spec not in combined_tabs[idx]['input_actions']:
                combined_tabs[idx]['input_actions'].append(input_spec)

        # add missing output actions
        for output_spec in tab_spec['output_actions']:

            action_spec = action_specs.get(output_spec)
            if not action_spec:
                raise Exception("Cannot read action " + output_spec + ".")

            if not include_eeg and 'eeg' in action_spec[2].get('tags', []):
                continue

            if output_spec not in combined_tabs[idx]['output_actions']:
                combined_tabs[idx]['output_actions'].append(output_spec)

        # add missing info elements
        for info_spec in tab_spec['info']:

            action_spec = action_specs.get(info_spec)
            if not action_spec:
                raise Exception("Cannot read info item " + info_spec + ".")

            if info_spec not in combined_tabs[idx]['info']:
                combined_tabs[idx]['info'].append(info_spec)

    # If everything is fine, construct each of the tabs
    tabs = []
    for tab_spec in combined_tabs:
        tabs.append(construct_tab(tab_spec, action_specs, datatype_specs, window))

    return tabs


def subject_action(inner_function):
    """ Decorator for automatically logging actions for each subject
    separately. Should be used within Action-instances.
    """
    def outer_function(self, subject, params, *args, **kwargs):

        message_dict = {
            'uid': self.uid,
            'type': 'SUBJECT_START',
            'subject_uid': subject.uid,
            'params': params
        }
        logging.getLogger('action_logger').info(message_dict)

        res = inner_function(self, subject, params, *args, **kwargs)

        message_dict = {
            'uid': self.uid,
            'type': 'SUBJECT_END',
            'subject_uid': subject.uid
        }
        logging.getLogger('action_logger').info(message_dict)

        message = '"{0}" finished successfully for subject {1}.'.format(
            self.action_spec['name'], subject.name)
        logging.getLogger('ui_logger').info(message)

        return res
    return outer_function


class Action:
    """ All actions should inherit this, so that automatic logging
    works.
    """
    logged = True

    def __init__(self, experiment, data, window, action_spec):
        """
        """
        self.experiment = experiment
        self.data = data
        self.window = window
        self.action_spec = action_spec

        # generate short temporary uid
        self.uid = generate_uid()

        if self.logged:
            # and log action with it
            message_dict = {
                'uid': self.uid,
                'type': 'ACTION_START',
                'id': action_spec['id'],
                'desc': self.data['tab_id']
            }
            logging.getLogger('action_logger').info(message_dict)

    def run(self):
        """ Called when action is initiated
        """
        messagebox("Running action " + self.action_spec['id'])


class InfoAction(Action):
    """ This should be inherited instead of Action
    when filling info boxes. """
    logged = False

    def run(self):
        """ Called when the info box is filled.
        """
        return "Returned content."




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'addSubjectDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AddSubject(object):
    def setupUi(self, AddSubject):
        AddSubject.setObjectName("AddSubject")
        AddSubject.resize(519, 312)
        self.gridLayout = QtWidgets.QGridLayout(AddSubject)
        self.gridLayout.setObjectName("gridLayout")
        self.labelAddSubjects = QtWidgets.QLabel(AddSubject)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.labelAddSubjects.setFont(font)
        self.labelAddSubjects.setObjectName("labelAddSubjects")
        self.gridLayout.addWidget(self.labelAddSubjects, 0, 0, 1, 2)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.pushButtonCancel = QtWidgets.QPushButton(AddSubject)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout_2.addWidget(self.pushButtonCancel)
        self.pushButtonOk = QtWidgets.QPushButton(AddSubject)
        self.pushButtonOk.setObjectName("pushButtonOk")
        self.horizontalLayout_2.addWidget(self.pushButtonOk)
        self.gridLayout.addLayout(self.horizontalLayout_2, 2, 0, 1, 3)
        self.scrollArea = QtWidgets.QScrollArea(AddSubject)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 499, 240))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.pushButtonRemove = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.pushButtonRemove.setEnabled(False)
        self.pushButtonRemove.setObjectName("pushButtonRemove")
        self.gridLayout_2.addWidget(self.pushButtonRemove, 2, 0, 1, 1)
        self.listWidgetFileNames = QtWidgets.QListWidget(self.scrollAreaWidgetContents)
        self.listWidgetFileNames.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.listWidgetFileNames.setObjectName("listWidgetFileNames")
        self.gridLayout_2.addWidget(self.listWidgetFileNames, 0, 0, 1, 1)
        self.pushButtonBrowse = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButtonBrowse.sizePolicy().hasHeightForWidth())
        self.pushButtonBrowse.setSizePolicy(sizePolicy)
        self.pushButtonBrowse.setObjectName("pushButtonBrowse")
        self.gridLayout_2.addWidget(self.pushButtonBrowse, 1, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 1, 0, 1, 3)

        self.retranslateUi(AddSubject)
        self.pushButtonCancel.clicked.connect(AddSubject.reject)
        self.pushButtonOk.clicked.connect(AddSubject.accept)
        QtCore.QMetaObject.connectSlotsByName(AddSubject)

    def retranslateUi(self, AddSubject):
        _translate = QtCore.QCoreApplication.translate
        AddSubject.setWindowTitle(_translate("AddSubject", "Meggie - Add subjects"))
        self.labelAddSubjects.setText(_translate("AddSubject", "Add subjects (raw files) to the experiment:"))
        self.pushButtonCancel.setText(_translate("AddSubject", "Cancel"))
        self.pushButtonOk.setText(_translate("AddSubject", "Ok"))
        self.pushButtonRemove.setText(_translate("AddSubject", "Remove"))
        self.pushButtonBrowse.setText(_translate("AddSubject", "Browse..."))




""" Contains a class for logic of events from annotations dialog.
"""
import logging

from PyQt5 import QtWidgets

from meggie.actions.raw_events_from_annotations.dialogs.eventsFromAnnotationsDialogUi import Ui_EventsFromAnnotationsDialog
from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.utilities.messaging import exc_messagebox


class EventsFromAnnotationsDialog(QtWidgets.QDialog):
    """ Contains logic for events from annotations dialog.
    """

    def __init__(self, parent, experiment, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_EventsFromAnnotationsDialog()
        self.ui.setupUi(self)

        self.experiment = experiment
        self.parent = parent
        self.handler = handler

        self.items = []

        subject = self.experiment.active_subject
        raw = subject.get_raw()

        for annotation_name in sorted(list(set(raw.annotations.description))):
            self.ui.comboBoxAnnotation.addItem(annotation_name)

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

    def _experiment_getter(self):
        return self.experiment

    def on_pushButtonAdd_clicked(self, checked=None):
        if checked is None:
            return

        annotation_name = self.ui.comboBoxAnnotation.currentText()
        event_id = self.ui.spinBoxEventID.value()
        use_start = True if self.ui.radioButtonStart.isChecked() else False

        self.items.append((annotation_name, event_id, use_start))
        self._update_list()

    def on_pushButtonClear_clicked(self, checked=None):
        if checked is None:
            return

        self.items = []
        self._update_list()

    def _update_list(self):
        self.ui.listWidgetItems.clear()

        for item in self.items:
            text = 'Annotation: ' + str(item[0]) + ', event id: ' + str(item[1]) + ', '
            if item[2]:
                text = text + 'use start'
            else:
                text = text + 'use end'

            item = QtWidgets.QListWidgetItem(str(text))
            self.ui.listWidgetItems.addItem(item)

    def accept(self):
        subject = self.experiment.active_subject

        try: 
            params = {'items': self.items}
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):
        experiment = self.experiment

        selected_subject_names = self.batching_widget.selected_subjects

        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    params = {'items': self.items}
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        self.parent.initialize_ui()
        self.close()



""" Contains a class for logic of the evoked topomap dialog.
"""
from PyQt5 import QtWidgets

from meggie.actions.evoked_plot_topomap.dialogs.evokedTopomapDialogUi import Ui_evokedTopomapDialog

from meggie.utilities.messaging import exc_messagebox


class EvokedTopomapDialog(QtWidgets.QDialog):
    """ Contains logic for the evoked topomap dialog.
    """
    def __init__(self, parent, evoked, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_evokedTopomapDialog()
        self.ui.setupUi(self)
        self.handler = handler
        self.evoked = evoked

        times = list(evoked.content.values())[0].times
        tmin, tmax = times[0], times[-1]

        self.ui.doubleSpinBoxStart.setValue(tmin)
        self.ui.doubleSpinBoxStart.setMinimum(tmin)
        self.ui.doubleSpinBoxStart.setMaximum(tmax)
        self.ui.doubleSpinBoxEnd.setMinimum(tmin)
        self.ui.doubleSpinBoxEnd.setMaximum(tmax)
        self.ui.doubleSpinBoxEnd.setValue(tmax)

    def accept(self):
        tmin = self.ui.doubleSpinBoxStart.value()
        tmax = self.ui.doubleSpinBoxEnd.value()
        step = self.ui.doubleSpinBoxStep.value()

        radius = None
        if self.ui.checkBoxRadius.isChecked():
            radius = self.ui.doubleSpinBoxRadius.value()

        try:
            self.handler(tmin, tmax, step, radius)
        except Exception as exc:
            exc_messagebox(self.parent, exc)

        self.close()




import os 

import mne
import numpy as np

from meggie.utilities.measurement_info import MeasurementInfo


def test_measurement_info():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, 'MEG', 'sample', 'sample_audvis_raw.fif')
    raw = mne.io.read_raw_fif(sample_fname, preload=True)

    meas_info = MeasurementInfo(raw)

    assert(meas_info.high_pass == 0.1)
    assert(meas_info.low_pass == 172.18)
    assert(meas_info.sampling_freq == 600.61)
    assert(meas_info.date == '2002-12-03')
    assert(meas_info.subject_name == '')




""" Contains implementation for epochs info
"""
from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float


class Info(InfoAction):
    """ Fills up epochs info box """

    def run(self):

        try:
            selected_name = self.data['outputs']['epochs'][0]

            epochs = self.experiment.active_subject.epochs[selected_name]
            params = epochs.params

            message = ""

            message += "Name: {0}\n\n".format(epochs.name)

            message += "Count: {0}\n".format(epochs.count)

            message += 'Baseline: {0}s - {1}s\n'.format(format_float(params['bstart']),
                                                        format_float(params['bend']))
            message += 'Times: {0}s - {1}s\n'.format(format_float(params['tmin']),
                                                     format_float(params['tmax']))

            message += '\nReject params: \n'
            if 'grad' in params:
                message += 'Gradiometers: {}\n'.format(params.get('reject').get('grad'))
            if 'mag' in params:
                message += 'Magnetometers: {}\n'.format(params.get('reject').get('mag'))
            if 'eeg' in params:
                message += 'EEG: {}\n'.format(params.get('reject').get('eeg'))

            if 'events' in params:
                message += '\nCreated from events: \n'
                for event in params.get('events'):
                    message += 'ID: {0}, mask: {1}\n'.format(event['event_id'],
                                                             event['mask'])

        except Exception as exc:
            message = ""

        return message





""" Contains implementation for delete spectrum
"""
import logging

import numpy as np

from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class DeleteSpectrum(Action):
    """ Deletes spectrum of selected name
    """

    def run(self):

        subject = self.experiment.active_subject

        try:
            selected_name = self.data['outputs']['spectrum'][0]
        except IndexError as exc:
            return

        try:
            self.handler(subject, {'name': selected_name})
        except Exception as exc:
            exc_messagebox(self.window, exc)
            self.experiment.save_experiment_settings()

        self.window.initialize_ui()

    @subject_action
    def handler(self, subject, params):
        subject.remove(params['name'], 'spectrum')





""" Contains controlling logic for the tfr implementation
"""

import os
import logging

import mne

import meggie.utilities.filemanager as filemanager

from meggie.utilities.formats import format_floats
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.threading import threaded


@threaded
def save_tfr_all_channels(experiment, tfr_name,
                          blmode, blstart, blend,
                          tmin, tmax, fmin, fmax):
    """ Saves all channels of tfr item to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        tfr = subject.tfr.get(tfr_name)
        if not tfr:
            continue

        ch_names = tfr.ch_names

        for key, mne_tfr in tfr.content.items():

            # crop and correct to baseline
            mne_tfr = mne_tfr.copy().crop(tmin=tmin,
                                          tmax=tmax,
                                          fmin=fmin,
                                          fmax=fmax)
            times = mne_tfr.times
            column_names = format_floats(times)
            freqs = format_floats(mne_tfr.freqs)

            data = mne.baseline.rescale(mne_tfr.data, times, baseline=(blstart, blend), 
                                       mode=blmode)

            for ix in range(data.shape[0]):
                for iy in range(data.shape[1]):
                    csv_data.append(data[ix, iy].tolist())

                    row_desc = (subject.name, key, ch_names[ix], freqs[iy])
                    row_descs.append(row_desc)

        folder = filemanager.create_timestamped_folder(experiment)
        fname = tfr_name + '_all_subjects_all_channels_tfr.csv'
        path = os.path.join(folder, fname)

        filemanager.save_csv(path, csv_data, column_names, row_descs)
        logging.getLogger('ui_logger').info('Saved the csv file to ' + path)

@threaded
def save_tfr_channel_averages(experiment, tfr_name,
                              blmode, blstart, blend,
                              tmin, tmax, fmin, fmax,
                              channel_groups):
    """ Saves channel averages of tfr item to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        tfr = subject.tfr.get(tfr_name)
        if not tfr:
            continue

        ch_names = tfr.ch_names
        info = tfr.info

        for key, mne_tfr in tfr.content.items():

            # crop and correct to baseline
            mne_tfr = mne_tfr.copy().crop(tmin=tmin,
                                          tmax=tmax,
                                          fmin=fmin,
                                          fmax=fmax)
            times = mne_tfr.times
            column_names = format_floats(times)
            freqs = format_floats(mne_tfr.freqs)

            data = mne.baseline.rescale(mne_tfr.data, times, baseline=(blstart, blend), 
                                       mode=blmode)

            data_labels, averaged_data = average_to_channel_groups(
                data, info, ch_names, channel_groups)

            for ix in range(averaged_data.shape[0]):
                for iy in range(averaged_data.shape[1]):
                    ch_type, area = data_labels[ix]

                    csv_data.append(averaged_data[ix, iy].tolist())

                    row_desc = (subject.name, key, ch_type, area, freqs[iy])
                    row_descs.append(row_desc)

        folder = filemanager.create_timestamped_folder(experiment)
        fname = tfr_name + '_all_subjects_channel_averages_tfr.csv'
        path = os.path.join(folder, fname)

        filemanager.save_csv(path, csv_data, column_names, row_descs)
        logging.getLogger('ui_logger').info('Saved the csv file to ' + path)





"""Provides for a class that wraps measurement information from the mne.Info.
"""

class MeasurementInfo(object):
    """A class for collecting information from raw files.

    Parameters
    ----------
    raw : mne.io.Raw
        The raw object the measurement info is based on.
    """

    def __init__(self, raw):
        self._info = raw.info

    @property
    def high_pass(self):
        """Returns the high-pass filter value.
        """
        try:
            return round(self._info.get('highpass'), 2)
        except Exception as exc:
            pass

    @property
    def low_pass(self):
        """Returns the low-pass filter value.
        """
        try:
            return round(self._info.get('lowpass'), 2)
        except Exception as exc:
            return ""

    @property
    def sampling_freq(self):
        """Returns the sampling frequency.
        """
        try:
            return round(self._info.get('sfreq'), 2)
        except Exception as exc:
            return ""

    @property
    def date(self):
        """Returns the measurement date."""
        try:
            return str(self._info['meas_date'].date())
        except Exception as exc:
            return ""

    @property
    def subject_name(self):
        """Constructs a one-string-name."""
        subj_info = self._info.get('subject_info')

        if not subj_info:
            return ''

        last_name = subj_info.get('last_name', '')
        first_name = subj_info.get('first_name', '')

        return last_name + ' ' + first_name




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer_ui_files/tabs/tfr/TFRDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TFRDialog(object):
    def setupUi(self, TFRDialog):
        TFRDialog.setObjectName("TFRDialog")
        TFRDialog.resize(512, 535)
        self.gridLayout_2 = QtWidgets.QGridLayout(TFRDialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.horizontalLayoutButtons = QtWidgets.QHBoxLayout()
        self.horizontalLayoutButtons.setObjectName("horizontalLayoutButtons")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayoutButtons.addItem(spacerItem)
        self.pushButtonCancel = QtWidgets.QPushButton(TFRDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayoutButtons.addWidget(self.pushButtonCancel)
        self.pushButtonBatch = QtWidgets.QPushButton(TFRDialog)
        self.pushButtonBatch.setObjectName("pushButtonBatch")
        self.horizontalLayoutButtons.addWidget(self.pushButtonBatch)
        self.pushButtonApply = QtWidgets.QPushButton(TFRDialog)
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.horizontalLayoutButtons.addWidget(self.pushButtonApply)
        self.gridLayout_2.addLayout(self.horizontalLayoutButtons, 1, 0, 1, 1)
        self.scrollArea = QtWidgets.QScrollArea(TFRDialog)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 480, 783))
        self.scrollAreaWidgetContents.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.groupBoxMisc = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxMisc.setObjectName("groupBoxMisc")
        self.gridLayout_9 = QtWidgets.QGridLayout(self.groupBoxMisc)
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.gridLayout_10 = QtWidgets.QGridLayout()
        self.gridLayout_10.setObjectName("gridLayout_10")
        self.spinBoxDecim = QtWidgets.QSpinBox(self.groupBoxMisc)
        self.spinBoxDecim.setMinimum(1)
        self.spinBoxDecim.setMaximum(1000)
        self.spinBoxDecim.setProperty("value", 1)
        self.spinBoxDecim.setObjectName("spinBoxDecim")
        self.gridLayout_10.addWidget(self.spinBoxDecim, 0, 1, 1, 1)
        self.labelDecim = QtWidgets.QLabel(self.groupBoxMisc)
        self.labelDecim.setObjectName("labelDecim")
        self.gridLayout_10.addWidget(self.labelDecim, 0, 0, 1, 1)
        self.gridLayout_9.addLayout(self.gridLayout_10, 1, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelSubtractEvoked = QtWidgets.QLabel(self.groupBoxMisc)
        self.labelSubtractEvoked.setObjectName("labelSubtractEvoked")
        self.horizontalLayout.addWidget(self.labelSubtractEvoked)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.checkBoxSubtractEvoked = QtWidgets.QCheckBox(self.groupBoxMisc)
        self.checkBoxSubtractEvoked.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.checkBoxSubtractEvoked.setText("")
        self.checkBoxSubtractEvoked.setObjectName("checkBoxSubtractEvoked")
        self.horizontalLayout.addWidget(self.checkBoxSubtractEvoked)
        self.gridLayout_9.addLayout(self.horizontalLayout, 2, 0, 1, 1)
        self.gridLayout_4.addWidget(self.groupBoxMisc, 4, 0, 1, 1)
        self.groupBoxFrequencies = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(2)
        sizePolicy.setHeightForWidth(self.groupBoxFrequencies.sizePolicy().hasHeightForWidth())
        self.groupBoxFrequencies.setSizePolicy(sizePolicy)
        self.groupBoxFrequencies.setObjectName("groupBoxFrequencies")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.groupBoxFrequencies)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.labelMinFreq = QtWidgets.QLabel(self.groupBoxFrequencies)
        self.labelMinFreq.setObjectName("labelMinFreq")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelMinFreq)
        self.doubleSpinBoxMinFreq = QtWidgets.QDoubleSpinBox(self.groupBoxFrequencies)
        self.doubleSpinBoxMinFreq.setMinimum(0.0)
        self.doubleSpinBoxMinFreq.setMaximum(1000.0)
        self.doubleSpinBoxMinFreq.setProperty("value", 5.0)
        self.doubleSpinBoxMinFreq.setObjectName("doubleSpinBoxMinFreq")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxMinFreq)
        self.labelMaxFreq = QtWidgets.QLabel(self.groupBoxFrequencies)
        self.labelMaxFreq.setObjectName("labelMaxFreq")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.labelMaxFreq)
        self.doubleSpinBoxMaxFreq = QtWidgets.QDoubleSpinBox(self.groupBoxFrequencies)
        self.doubleSpinBoxMaxFreq.setMinimum(0.0)
        self.doubleSpinBoxMaxFreq.setMaximum(1000.0)
        self.doubleSpinBoxMaxFreq.setProperty("value", 30.0)
        self.doubleSpinBoxMaxFreq.setObjectName("doubleSpinBoxMaxFreq")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxMaxFreq)
        self.labelFrequencyInterval = QtWidgets.QLabel(self.groupBoxFrequencies)
        self.labelFrequencyInterval.setObjectName("labelFrequencyInterval")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.labelFrequencyInterval)
        self.doubleSpinBoxFreqInterval = QtWidgets.QDoubleSpinBox(self.groupBoxFrequencies)
        self.doubleSpinBoxFreqInterval.setMinimum(0.1)
        self.doubleSpinBoxFreqInterval.setMaximum(99.99)
        self.doubleSpinBoxFreqInterval.setProperty("value", 0.5)
        self.doubleSpinBoxFreqInterval.setObjectName("doubleSpinBoxFreqInterval")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxFreqInterval)
        self.label_7 = QtWidgets.QLabel(self.groupBoxFrequencies)
        font = QtGui.QFont()
        font.setBold(True)
        font.setItalic(False)
        font.setUnderline(False)
        font.setWeight(75)
        font.setStrikeOut(False)
        self.label_7.setFont(font)
        self.label_7.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.horizontalLayout_18 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_18.setObjectName("horizontalLayout_18")
        self.radioButtonFixed = QtWidgets.QRadioButton(self.groupBoxFrequencies)
        self.radioButtonFixed.setChecked(False)
        self.radioButtonFixed.setObjectName("radioButtonFixed")
        self.horizontalLayout_18.addWidget(self.radioButtonFixed)
        self.formLayout.setLayout(4, QtWidgets.QFormLayout.LabelRole, self.horizontalLayout_18)
        self.doubleSpinBoxNcycles = QtWidgets.QDoubleSpinBox(self.groupBoxFrequencies)
        self.doubleSpinBoxNcycles.setEnabled(False)
        self.doubleSpinBoxNcycles.setMinimum(1.0)
        self.doubleSpinBoxNcycles.setMaximum(100.0)
        self.doubleSpinBoxNcycles.setProperty("value", 5.0)
        self.doubleSpinBoxNcycles.setObjectName("doubleSpinBoxNcycles")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxNcycles)
        self.horizontalLayout_19 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_19.setObjectName("horizontalLayout_19")
        self.radioButtonAdapted = QtWidgets.QRadioButton(self.groupBoxFrequencies)
        self.radioButtonAdapted.setToolTip("")
        self.radioButtonAdapted.setCheckable(True)
        self.radioButtonAdapted.setChecked(True)
        self.radioButtonAdapted.setObjectName("radioButtonAdapted")
        self.horizontalLayout_19.addWidget(self.radioButtonAdapted)
        self.formLayout.setLayout(5, QtWidgets.QFormLayout.LabelRole, self.horizontalLayout_19)
        self.doubleSpinBoxCycleFactor = QtWidgets.QDoubleSpinBox(self.groupBoxFrequencies)
        self.doubleSpinBoxCycleFactor.setEnabled(True)
        self.doubleSpinBoxCycleFactor.setToolTip("")
        self.doubleSpinBoxCycleFactor.setMinimum(0.0)
        self.doubleSpinBoxCycleFactor.setMaximum(10.0)
        self.doubleSpinBoxCycleFactor.setSingleStep(1.0)
        self.doubleSpinBoxCycleFactor.setProperty("value", 2.0)
        self.doubleSpinBoxCycleFactor.setObjectName("doubleSpinBoxCycleFactor")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxCycleFactor)
        self.gridLayout_5.addLayout(self.formLayout, 0, 1, 1, 1)
        self.gridLayout_4.addWidget(self.groupBoxFrequencies, 3, 0, 1, 2)
        self.groupBoxGeneral = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxGeneral.setObjectName("groupBoxGeneral")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.groupBoxGeneral)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.labelEpochName = QtWidgets.QLabel(self.groupBoxGeneral)
        self.labelEpochName.setObjectName("labelEpochName")
        self.gridLayout_6.addWidget(self.labelEpochName, 1, 0, 1, 1)
        self.lineEditEpochName = QtWidgets.QLineEdit(self.groupBoxGeneral)
        self.lineEditEpochName.setEnabled(False)
        self.lineEditEpochName.setObjectName("lineEditEpochName")
        self.gridLayout_6.addWidget(self.lineEditEpochName, 1, 1, 1, 1)
        self.labelTFRName = QtWidgets.QLabel(self.groupBoxGeneral)
        self.labelTFRName.setObjectName("labelTFRName")
        self.gridLayout_6.addWidget(self.labelTFRName, 2, 0, 1, 1)
        self.lineEditTFRName = QtWidgets.QLineEdit(self.groupBoxGeneral)
        self.lineEditTFRName.setObjectName("lineEditTFRName")
        self.gridLayout_6.addWidget(self.lineEditTFRName, 2, 1, 1, 1)
        self.gridLayout_4.addWidget(self.groupBoxGeneral, 0, 0, 1, 1)
        self.groupBoxBatching = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxBatching.setObjectName("groupBoxBatching")
        self.gridLayoutBatching = QtWidgets.QGridLayout(self.groupBoxBatching)
        self.gridLayoutBatching.setObjectName("gridLayoutBatching")
        self.batchingWidgetPlaceholder = QtWidgets.QWidget(self.groupBoxBatching)
        self.batchingWidgetPlaceholder.setMinimumSize(QtCore.QSize(300, 300))
        self.batchingWidgetPlaceholder.setObjectName("batchingWidgetPlaceholder")
        self.gridLayoutBatching.addWidget(self.batchingWidgetPlaceholder, 0, 0, 1, 1)
        self.gridLayout_4.addWidget(self.groupBoxBatching, 5, 0, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_4.addItem(spacerItem2, 6, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout_2.addWidget(self.scrollArea, 0, 0, 1, 1)

        self.retranslateUi(TFRDialog)
        self.radioButtonAdapted.toggled['bool'].connect(self.doubleSpinBoxCycleFactor.setEnabled)
        self.radioButtonFixed.toggled['bool'].connect(self.doubleSpinBoxNcycles.setEnabled)
        self.pushButtonCancel.clicked.connect(TFRDialog.reject)
        self.pushButtonBatch.clicked.connect(TFRDialog.acceptBatch)
        self.pushButtonApply.clicked.connect(TFRDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(TFRDialog)
        TFRDialog.setTabOrder(self.scrollArea, self.radioButtonFixed)
        TFRDialog.setTabOrder(self.radioButtonFixed, self.radioButtonAdapted)

    def retranslateUi(self, TFRDialog):
        _translate = QtCore.QCoreApplication.translate
        TFRDialog.setWindowTitle(_translate("TFRDialog", "Meggie - Compute TFR"))
        self.pushButtonCancel.setText(_translate("TFRDialog", "Cancel"))
        self.pushButtonBatch.setText(_translate("TFRDialog", "Batch"))
        self.pushButtonApply.setText(_translate("TFRDialog", "Apply"))
        self.groupBoxMisc.setTitle(_translate("TFRDialog", "Miscellaneous"))
        self.labelDecim.setText(_translate("TFRDialog", "Decimation factor:"))
        self.labelSubtractEvoked.setText(_translate("TFRDialog", "Subtract evoked:"))
        self.groupBoxFrequencies.setTitle(_translate("TFRDialog", "Frequency window"))
        self.labelMinFreq.setText(_translate("TFRDialog", "Min frequency:"))
        self.doubleSpinBoxMinFreq.setSuffix(_translate("TFRDialog", "Hz"))
        self.labelMaxFreq.setText(_translate("TFRDialog", "Max frequency:"))
        self.doubleSpinBoxMaxFreq.setSuffix(_translate("TFRDialog", "Hz"))
        self.labelFrequencyInterval.setText(_translate("TFRDialog", "Frequency interval:"))
        self.doubleSpinBoxFreqInterval.setSuffix(_translate("TFRDialog", "Hz"))
        self.label_7.setText(_translate("TFRDialog", "Number of cycles"))
        self.radioButtonFixed.setText(_translate("TFRDialog", "Fixed"))
        self.radioButtonAdapted.setText(_translate("TFRDialog", "Freqs divided by"))
        self.groupBoxGeneral.setTitle(_translate("TFRDialog", "General:"))
        self.labelEpochName.setText(_translate("TFRDialog", "Epoch name: "))
        self.labelTFRName.setText(_translate("TFRDialog", "TFR name:"))
        self.groupBoxBatching.setTitle(_translate("TFRDialog", "Batching"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'bitSelectionDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(919, 175)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.labelID = QtWidgets.QLabel(Dialog)
        self.labelID.setText("")
        self.labelID.setObjectName("labelID")
        self.horizontalLayout_3.addWidget(self.labelID)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.labelMask = QtWidgets.QLabel(Dialog)
        self.labelMask.setText("")
        self.labelMask.setObjectName("labelMask")
        self.horizontalLayout_3.addWidget(self.labelMask)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout16 = QtWidgets.QVBoxLayout()
        self.verticalLayout16.setObjectName("verticalLayout16")
        self.pushButton16 = QtWidgets.QPushButton(Dialog)
        self.pushButton16.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton16.setText("")
        self.pushButton16.setObjectName("pushButton16")
        self.verticalLayout16.addWidget(self.pushButton16)
        self.bitLabel16 = QtWidgets.QLabel(Dialog)
        self.bitLabel16.setMaximumSize(QtCore.QSize(40, 16777215))
        self.bitLabel16.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel16.setObjectName("bitLabel16")
        self.verticalLayout16.addWidget(self.bitLabel16)
        self.horizontalLayout.addLayout(self.verticalLayout16)
        self.line16 = QtWidgets.QFrame(Dialog)
        self.line16.setFrameShape(QtWidgets.QFrame.VLine)
        self.line16.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line16.setObjectName("line16")
        self.horizontalLayout.addWidget(self.line16)
        self.verticalLayout15 = QtWidgets.QVBoxLayout()
        self.verticalLayout15.setObjectName("verticalLayout15")
        self.pushButton15 = QtWidgets.QPushButton(Dialog)
        self.pushButton15.setMinimumSize(QtCore.QSize(0, 0))
        self.pushButton15.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton15.setText("")
        self.pushButton15.setObjectName("pushButton15")
        self.verticalLayout15.addWidget(self.pushButton15)
        self.bitLabel15 = QtWidgets.QLabel(Dialog)
        self.bitLabel15.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel15.setObjectName("bitLabel15")
        self.verticalLayout15.addWidget(self.bitLabel15)
        self.horizontalLayout.addLayout(self.verticalLayout15)
        self.line15 = QtWidgets.QFrame(Dialog)
        self.line15.setFrameShape(QtWidgets.QFrame.VLine)
        self.line15.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line15.setObjectName("line15")
        self.horizontalLayout.addWidget(self.line15)
        self.verticalLayout14 = QtWidgets.QVBoxLayout()
        self.verticalLayout14.setObjectName("verticalLayout14")
        self.pushButton14 = QtWidgets.QPushButton(Dialog)
        self.pushButton14.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton14.setText("")
        self.pushButton14.setObjectName("pushButton14")
        self.verticalLayout14.addWidget(self.pushButton14)
        self.bitLabel14 = QtWidgets.QLabel(Dialog)
        self.bitLabel14.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel14.setObjectName("bitLabel14")
        self.verticalLayout14.addWidget(self.bitLabel14)
        self.horizontalLayout.addLayout(self.verticalLayout14)
        self.line14 = QtWidgets.QFrame(Dialog)
        self.line14.setFrameShape(QtWidgets.QFrame.VLine)
        self.line14.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line14.setObjectName("line14")
        self.horizontalLayout.addWidget(self.line14)
        self.verticalLayout13 = QtWidgets.QVBoxLayout()
        self.verticalLayout13.setObjectName("verticalLayout13")
        self.pushButton13 = QtWidgets.QPushButton(Dialog)
        self.pushButton13.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton13.setText("")
        self.pushButton13.setObjectName("pushButton13")
        self.verticalLayout13.addWidget(self.pushButton13)
        self.bitLabel13 = QtWidgets.QLabel(Dialog)
        self.bitLabel13.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel13.setObjectName("bitLabel13")
        self.verticalLayout13.addWidget(self.bitLabel13)
        self.horizontalLayout.addLayout(self.verticalLayout13)
        self.line13 = QtWidgets.QFrame(Dialog)
        self.line13.setFrameShape(QtWidgets.QFrame.VLine)
        self.line13.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line13.setObjectName("line13")
        self.horizontalLayout.addWidget(self.line13)
        self.verticalLayout12 = QtWidgets.QVBoxLayout()
        self.verticalLayout12.setObjectName("verticalLayout12")
        self.pushButton12 = QtWidgets.QPushButton(Dialog)
        self.pushButton12.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton12.setText("")
        self.pushButton12.setObjectName("pushButton12")
        self.verticalLayout12.addWidget(self.pushButton12)
        self.bitLabel12 = QtWidgets.QLabel(Dialog)
        self.bitLabel12.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel12.setObjectName("bitLabel12")
        self.verticalLayout12.addWidget(self.bitLabel12)
        self.horizontalLayout.addLayout(self.verticalLayout12)
        self.line12 = QtWidgets.QFrame(Dialog)
        self.line12.setFrameShape(QtWidgets.QFrame.VLine)
        self.line12.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line12.setObjectName("line12")
        self.horizontalLayout.addWidget(self.line12)
        self.verticalLayout11 = QtWidgets.QVBoxLayout()
        self.verticalLayout11.setObjectName("verticalLayout11")
        self.pushButton11 = QtWidgets.QPushButton(Dialog)
        self.pushButton11.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton11.setText("")
        self.pushButton11.setObjectName("pushButton11")
        self.verticalLayout11.addWidget(self.pushButton11)
        self.bitLabel11 = QtWidgets.QLabel(Dialog)
        self.bitLabel11.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel11.setObjectName("bitLabel11")
        self.verticalLayout11.addWidget(self.bitLabel11)
        self.horizontalLayout.addLayout(self.verticalLayout11)
        self.line11 = QtWidgets.QFrame(Dialog)
        self.line11.setFrameShape(QtWidgets.QFrame.VLine)
        self.line11.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line11.setObjectName("line11")
        self.horizontalLayout.addWidget(self.line11)
        self.verticalLayout10 = QtWidgets.QVBoxLayout()
        self.verticalLayout10.setObjectName("verticalLayout10")
        self.pushButton10 = QtWidgets.QPushButton(Dialog)
        self.pushButton10.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton10.setText("")
        self.pushButton10.setObjectName("pushButton10")
        self.verticalLayout10.addWidget(self.pushButton10)
        self.bitLabel10 = QtWidgets.QLabel(Dialog)
        self.bitLabel10.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel10.setObjectName("bitLabel10")
        self.verticalLayout10.addWidget(self.bitLabel10)
        self.horizontalLayout.addLayout(self.verticalLayout10)
        self.line10 = QtWidgets.QFrame(Dialog)
        self.line10.setFrameShape(QtWidgets.QFrame.VLine)
        self.line10.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line10.setObjectName("line10")
        self.horizontalLayout.addWidget(self.line10)
        self.verticalLayout9 = QtWidgets.QVBoxLayout()
        self.verticalLayout9.setObjectName("verticalLayout9")
        self.pushButton9 = QtWidgets.QPushButton(Dialog)
        self.pushButton9.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton9.setText("")
        self.pushButton9.setObjectName("pushButton9")
        self.verticalLayout9.addWidget(self.pushButton9)
        self.bitLabel9 = QtWidgets.QLabel(Dialog)
        self.bitLabel9.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel9.setObjectName("bitLabel9")
        self.verticalLayout9.addWidget(self.bitLabel9)
        self.horizontalLayout.addLayout(self.verticalLayout9)
        self.line9 = QtWidgets.QFrame(Dialog)
        self.line9.setFrameShape(QtWidgets.QFrame.VLine)
        self.line9.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line9.setObjectName("line9")
        self.horizontalLayout.addWidget(self.line9)
        self.verticalLayout8 = QtWidgets.QVBoxLayout()
        self.verticalLayout8.setObjectName("verticalLayout8")
        self.pushButton8 = QtWidgets.QPushButton(Dialog)
        self.pushButton8.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton8.setText("")
        self.pushButton8.setObjectName("pushButton8")
        self.verticalLayout8.addWidget(self.pushButton8)
        self.bitLabel8 = QtWidgets.QLabel(Dialog)
        self.bitLabel8.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel8.setObjectName("bitLabel8")
        self.verticalLayout8.addWidget(self.bitLabel8)
        self.horizontalLayout.addLayout(self.verticalLayout8)
        self.line8 = QtWidgets.QFrame(Dialog)
        self.line8.setFrameShape(QtWidgets.QFrame.VLine)
        self.line8.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line8.setObjectName("line8")
        self.horizontalLayout.addWidget(self.line8)
        self.verticalLayout7 = QtWidgets.QVBoxLayout()
        self.verticalLayout7.setObjectName("verticalLayout7")
        self.pushButton7 = QtWidgets.QPushButton(Dialog)
        self.pushButton7.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton7.setText("")
        self.pushButton7.setObjectName("pushButton7")
        self.verticalLayout7.addWidget(self.pushButton7)
        self.bitLabel7 = QtWidgets.QLabel(Dialog)
        self.bitLabel7.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel7.setObjectName("bitLabel7")
        self.verticalLayout7.addWidget(self.bitLabel7)
        self.horizontalLayout.addLayout(self.verticalLayout7)
        self.line7 = QtWidgets.QFrame(Dialog)
        self.line7.setFrameShape(QtWidgets.QFrame.VLine)
        self.line7.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line7.setObjectName("line7")
        self.horizontalLayout.addWidget(self.line7)
        self.verticalLayout6 = QtWidgets.QVBoxLayout()
        self.verticalLayout6.setObjectName("verticalLayout6")
        self.pushButton6 = QtWidgets.QPushButton(Dialog)
        self.pushButton6.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton6.setText("")
        self.pushButton6.setObjectName("pushButton6")
        self.verticalLayout6.addWidget(self.pushButton6)
        self.bitLabel6 = QtWidgets.QLabel(Dialog)
        self.bitLabel6.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel6.setObjectName("bitLabel6")
        self.verticalLayout6.addWidget(self.bitLabel6)
        self.horizontalLayout.addLayout(self.verticalLayout6)
        self.line6 = QtWidgets.QFrame(Dialog)
        self.line6.setFrameShape(QtWidgets.QFrame.VLine)
        self.line6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line6.setObjectName("line6")
        self.horizontalLayout.addWidget(self.line6)
        self.verticalLayout5 = QtWidgets.QVBoxLayout()
        self.verticalLayout5.setObjectName("verticalLayout5")
        self.pushButton5 = QtWidgets.QPushButton(Dialog)
        self.pushButton5.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton5.setText("")
        self.pushButton5.setObjectName("pushButton5")
        self.verticalLayout5.addWidget(self.pushButton5)
        self.bitLabel5 = QtWidgets.QLabel(Dialog)
        self.bitLabel5.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel5.setObjectName("bitLabel5")
        self.verticalLayout5.addWidget(self.bitLabel5)
        self.horizontalLayout.addLayout(self.verticalLayout5)
        self.line5 = QtWidgets.QFrame(Dialog)
        self.line5.setFrameShape(QtWidgets.QFrame.VLine)
        self.line5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line5.setObjectName("line5")
        self.horizontalLayout.addWidget(self.line5)
        self.verticalLayout4 = QtWidgets.QVBoxLayout()
        self.verticalLayout4.setObjectName("verticalLayout4")
        self.pushButton4 = QtWidgets.QPushButton(Dialog)
        self.pushButton4.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton4.setText("")
        self.pushButton4.setObjectName("pushButton4")
        self.verticalLayout4.addWidget(self.pushButton4)
        self.bitLabel4 = QtWidgets.QLabel(Dialog)
        self.bitLabel4.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel4.setObjectName("bitLabel4")
        self.verticalLayout4.addWidget(self.bitLabel4)
        self.horizontalLayout.addLayout(self.verticalLayout4)
        self.line4 = QtWidgets.QFrame(Dialog)
        self.line4.setFrameShape(QtWidgets.QFrame.VLine)
        self.line4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line4.setObjectName("line4")
        self.horizontalLayout.addWidget(self.line4)
        self.verticalLayout3 = QtWidgets.QVBoxLayout()
        self.verticalLayout3.setObjectName("verticalLayout3")
        self.pushButton3 = QtWidgets.QPushButton(Dialog)
        self.pushButton3.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton3.setText("")
        self.pushButton3.setObjectName("pushButton3")
        self.verticalLayout3.addWidget(self.pushButton3)
        self.bitLabel3 = QtWidgets.QLabel(Dialog)
        self.bitLabel3.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel3.setObjectName("bitLabel3")
        self.verticalLayout3.addWidget(self.bitLabel3)
        self.horizontalLayout.addLayout(self.verticalLayout3)
        self.line3 = QtWidgets.QFrame(Dialog)
        self.line3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line3.setObjectName("line3")
        self.horizontalLayout.addWidget(self.line3)
        self.verticalLayout2 = QtWidgets.QVBoxLayout()
        self.verticalLayout2.setObjectName("verticalLayout2")
        self.pushButton2 = QtWidgets.QPushButton(Dialog)
        self.pushButton2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton2.setText("")
        self.pushButton2.setObjectName("pushButton2")
        self.verticalLayout2.addWidget(self.pushButton2)
        self.bitLabel2 = QtWidgets.QLabel(Dialog)
        self.bitLabel2.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel2.setObjectName("bitLabel2")
        self.verticalLayout2.addWidget(self.bitLabel2)
        self.horizontalLayout.addLayout(self.verticalLayout2)
        self.line2 = QtWidgets.QFrame(Dialog)
        self.line2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line2.setObjectName("line2")
        self.horizontalLayout.addWidget(self.line2)
        self.verticalLayout1 = QtWidgets.QVBoxLayout()
        self.verticalLayout1.setObjectName("verticalLayout1")
        self.pushButton1 = QtWidgets.QPushButton(Dialog)
        self.pushButton1.setMaximumSize(QtCore.QSize(40, 16777215))
        self.pushButton1.setText("")
        self.pushButton1.setObjectName("pushButton1")
        self.verticalLayout1.addWidget(self.pushButton1)
        self.bitLabel1 = QtWidgets.QLabel(Dialog)
        self.bitLabel1.setAlignment(QtCore.Qt.AlignCenter)
        self.bitLabel1.setObjectName("bitLabel1")
        self.verticalLayout1.addWidget(self.bitLabel1)
        self.horizontalLayout.addLayout(self.verticalLayout1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem3)
        self.pushButtonReject = QtWidgets.QPushButton(Dialog)
        self.pushButtonReject.setObjectName("pushButtonReject")
        self.horizontalLayout_2.addWidget(self.pushButtonReject)
        self.pushButtonAccept = QtWidgets.QPushButton(Dialog)
        self.pushButtonAccept.setObjectName("pushButtonAccept")
        self.horizontalLayout_2.addWidget(self.pushButtonAccept)
        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 0, 1, 1)

        self.retranslateUi(Dialog)
        self.pushButtonReject.clicked.connect(Dialog.reject)
        self.pushButtonAccept.clicked.connect(Dialog.accept)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Meggie - Edit"))
        self.label.setText(_translate("Dialog", "ID:"))
        self.label_3.setText(_translate("Dialog", "Mask:"))
        self.bitLabel16.setText(_translate("Dialog", "16"))
        self.bitLabel15.setText(_translate("Dialog", "15"))
        self.bitLabel14.setText(_translate("Dialog", "14"))
        self.bitLabel13.setText(_translate("Dialog", "13"))
        self.bitLabel12.setText(_translate("Dialog", "12"))
        self.bitLabel11.setText(_translate("Dialog", "11"))
        self.bitLabel10.setText(_translate("Dialog", "10"))
        self.bitLabel9.setText(_translate("Dialog", "9"))
        self.bitLabel8.setText(_translate("Dialog", "8"))
        self.bitLabel7.setText(_translate("Dialog", "7"))
        self.bitLabel6.setText(_translate("Dialog", "6"))
        self.bitLabel5.setText(_translate("Dialog", "5"))
        self.bitLabel4.setText(_translate("Dialog", "4"))
        self.bitLabel3.setText(_translate("Dialog", "3"))
        self.bitLabel2.setText(_translate("Dialog", "2"))
        self.bitLabel1.setText(_translate("Dialog", "1"))
        self.pushButtonReject.setText(_translate("Dialog", "Cancel"))
        self.pushButtonAccept.setText(_translate("Dialog", "Ok"))




""" Contains controlling logic for saving tse
"""

import os
import logging

import numpy as np
import mne

import meggie.utilities.filemanager as filemanager

from meggie.utilities.formats import format_floats
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.threading import threaded


def _compute_tse(meggie_tfr, fmin, fmax):
    tfrs = meggie_tfr.content

    fmin_idx = np.where(meggie_tfr.freqs >= fmin)[0][0]
    fmax_idx = np.where(meggie_tfr.freqs <= fmax)[0][-1]

    if fmax_idx <= fmin_idx:
        raise Exception('Something wrong with the frequencies')

    tses = {}
    for key, tfr in tfrs.items():
        tse = np.mean(tfr.data[:, fmin_idx:fmax_idx+1, :], axis=1)
        tses[key] = tse

    return tses


def _crop_and_correct_to_baseline(tse, blmode, blstart, blend, tmin, tmax, times):
    tmin_idx = np.where(times >= tmin)[0][0]
    tmax_idx = np.where(times <= tmax)[0][-1]

    if blmode:
        if blstart < tmin:
            raise Exception(
                'Baseline start should not be earlier than crop start.')

        if blend > tmax:
            raise Exception(
                'Baseline end should not be later than crop end.')

        # correct to baseline
        tse = mne.baseline.rescale(tse, times, baseline=(blstart, blend), 
                                   mode=blmode)

        # crop
        tse = tse[:, tmin_idx:tmax_idx+1]

    return times[tmin_idx:tmax_idx+1], tse


@threaded
def save_tse_all_channels(experiment, tfr_name, blmode, blstart, 
                          blend, tmin, tmax, fmin, fmax):
    """ Saves all channels of a tse to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        tfr = subject.tfr.get(tfr_name)
        if not tfr:
            continue

        tses = _compute_tse(tfr, fmin, fmax)

        for key, tse in tses.items():
            times, tse = _crop_and_correct_to_baseline(
                tse, blmode, blstart, blend, tmin, tmax, tfr.times)

            csv_data.extend(tse.tolist())

            for ch_name in tfr.ch_names:
                row_desc = (subject.name, key, ch_name)
                row_descs.append(row_desc)

        column_names = format_floats(times)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = tfr_name + '_all_subjects_all_channels_tse.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)


@threaded
def save_tse_channel_averages(experiment, tfr_name, blmode, blstart, 
                              blend, tmin, tmax, fmin, fmax,
                              channel_groups):
    """ Saves channel averages of tse to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        tfr = subject.tfr.get(tfr_name)
        if not tfr:
            continue

        tses = _compute_tse(tfr, fmin, fmax)

        for key, tse in tses.items():

            data_labels, averaged_data = average_to_channel_groups(
                tse, tfr.info, tfr.ch_names, channel_groups)

            times, averaged_data = _crop_and_correct_to_baseline(
                averaged_data, blmode, blstart, blend, tmin, tmax, tfr.times)

            csv_data.extend(averaged_data.tolist())

            for ch_type, area in data_labels:
                row_desc = (subject.name, key, ch_type, area)
                row_descs.append(row_desc)

        column_names = format_floats(times)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = tfr_name + '_all_subjects_channel_averages_tse.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)




""" Contains a class for logic of the group selection dialog.
"""
import logging

from PyQt5 import QtWidgets
from PyQt5 import QtCore

from meggie.utilities.dialogs.groupSelectionDialogUi import Ui_groupSelectionDialog

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.validators import validate_name


class GroupSelectionDialog(QtWidgets.QDialog):
    """ Contains the logic for group selection dialog.
    """

    def __init__(self, experiment, parent, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_groupSelectionDialog()
        self.ui.setupUi(self)

        self.handler = handler

        subjects = experiment.subjects.keys()
        subject_count = len(subjects)

        # fill the dialog with subjects
        for idx, subject_name in enumerate(subjects):
            self._add_item(idx, subject_name)

        self.subjects = subjects

    def _add_item(self, idx, name):
        setattr(self.ui, 'horizontalLayoutGroup_' +
                str(idx), QtWidgets.QHBoxLayout())
        getattr(self.ui, 'horizontalLayoutGroup_' + str(idx)).setObjectName(
            'horizontalLayoutGroup_' + str(idx))

        setattr(self.ui, 'checkBoxGroup_' + str(idx),
                QtWidgets.QCheckBox(self.ui.groupBoxGroups))
        getattr(self.ui, 'checkBoxGroup_' + str(idx)
                ).setObjectName('checkBoxGroup_' + str(idx))
        getattr(self.ui, 'checkBoxGroup_' + str(idx)).setText('')
        getattr(self.ui, 'checkBoxGroup_' + str(idx)).setMaximumSize(20, 20)
        getattr(self.ui, 'checkBoxGroup_' + str(idx)
                ).setCheckState(QtCore.Qt.Checked)
        getattr(self.ui, 'horizontalLayoutGroup_' + str(idx)).addWidget(
            getattr(self.ui, 'checkBoxGroup_' + str(idx)))

        setattr(self.ui, 'labelGroup_' + str(idx),
                QtWidgets.QLabel(self.ui.groupBoxGroups))
        getattr(self.ui, 'labelGroup_' + str(idx)
                ).setObjectName('labelGroup_' + str(idx))
        getattr(self.ui, 'labelGroup_' + str(idx)).setText(name)
        getattr(self.ui, 'horizontalLayoutGroup_' + str(idx)).addWidget(
            getattr(self.ui, 'labelGroup_' + str(idx)))

        setattr(self.ui, 'spinBoxGroup_' + str(idx),
                QtWidgets.QSpinBox(self.ui.groupBoxGroups))
        getattr(self.ui, 'spinBoxGroup_' + str(idx)).setMinimum(1)
        getattr(self.ui, 'spinBoxGroup_' + str(idx)).setMaximumSize(40, 1000)
        getattr(self.ui, 'spinBoxGroup_' + str(idx)
                ).setObjectName('spinBoxGroup_' + str(idx))
        getattr(self.ui, 'horizontalLayoutGroup_' + str(idx)).addWidget(
            getattr(self.ui, 'spinBoxGroup_' + str(idx)))

        self.ui.gridLayout.addLayout(getattr(self.ui, 'horizontalLayoutGroup_' + str(idx)),
                                     idx, 2, 1, 1)

    def accept(self):
        groups = {}
        for idx, subject in enumerate(self.subjects):
            selected = getattr(self.ui, 'checkBoxGroup_' +
                               str(idx)).checkState()
            if selected != QtCore.Qt.Checked:
                continue
            group_id = getattr(self.ui, 'spinBoxGroup_' + str(idx)).value()
            if group_id in groups:
                groups[group_id].append(subject)
            else:
                groups[group_id] = [subject]

        self.handler(groups)

        self.close()



""" Handles the logic of simple reusable dialogs
"""

import logging

import numpy as np

from PyQt5 import QtWidgets

from meggie.utilities.dialogs.simpleDialogUi import Ui_SimpleDialog

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.utilities.validators import validate_name
from meggie.utilities.validators import assert_arrays_same
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox


class SimpleDialog(QtWidgets.QDialog):
    """ Contains logic for simple reusable dialog.
    """

    def __init__(self, experiment, parent, default_name, handler, 
                 batching=True, title='Simple dialog'):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_SimpleDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment
        self.handler = handler

        self.setWindowTitle('Meggie - ' + title)

        if batching:
            self.batching_widget = BatchingWidget(
                experiment_getter=self._experiment_getter,
                parent=self,
                container=self.ui.groupBoxBatching,
                geometry=self.ui.batchingWidgetPlaceholder.geometry())
            self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)
        else:
            self.ui.groupBoxBatching.hide()
            self.ui.pushButtonBatch.hide()

        self.ui.lineEditName.setText(default_name)

    def _experiment_getter(self):
        return self.experiment

    def accept(self):
        subject = self.experiment.active_subject

        try:
            evoked_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self,  exc)
            return

        try:
            params = {'name': evoked_name}
            self.handler(subject, params)
            self.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):

        experiment = self.experiment

        try:
            evoked_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self,  exc)
            return

        selected_subject_names = self.batching_widget.selected_subjects

        params = {'name': evoked_name}

        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        try:
            self.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()



""" Contains implementation for evoked info
"""
from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float


class Info(InfoAction):
    """ Fills up evoked info box """

    def run(self):
        try:
            selected_name = self.data['outputs']['evoked'][0]
            evoked = self.experiment.active_subject.evoked[selected_name]
            params = evoked.params

            message = ""

            message += "Name: {}\n\n".format(evoked.name)

            if 'conditions' in params:
                message += 'Conditions: ' + ', '.join(params['conditions']) + '\n'

            if 'groups' in params:
                for key, names in params['groups'].items():
                    message += '\nGroup '+ str(key) + ': \n'
                    for name in names:
                        message += name + '\n'

        except Exception as exc:
            message = ""

        return message




""" Contains implementation for raw filter
"""
import logging

import matplotlib.pyplot as plt
import numpy as np

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.threading import threaded

from meggie.actions.raw_filter.dialogs.filterDialogMain import FilterDialog
from meggie.actions.raw_filter.controller.filter import filter_data

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class Filter(Action):
    """ Shows a dialog for getting parameters and allows
    filtering the raw data.
    """

    @subject_action
    def handler(self, subject, params):
        """
        """
        @threaded
        def filter_fun():
            filter_data(subject, params)

        filter_fun(do_meanwhile=self.window.update_ui)

    def run(self):
        filter_dialog = FilterDialog(
            self.window, self.experiment, self.handler)
        filter_dialog.show()



""" Contains controlling logic for the spectrum saving.
"""

import os
import logging

import mne
import numpy as np

import meggie.utilities.filemanager as filemanager

from meggie.utilities.formats import format_floats
from meggie.utilities.channels import average_to_channel_groups

from meggie.utilities.threading import threaded


@threaded
def save_all_channels(experiment, selected_name):
    """ Saves all channesl of a spectrum item to a csv file. """
    column_names = []
    row_descs = []
    csv_data = []

    for subject in experiment.subjects.values():
        spectrum = subject.spectrum.get(selected_name)
        if not spectrum:
            continue
        for key, psd in spectrum.content.items():
            csv_data.extend(psd.tolist())
            column_names = format_floats(spectrum.freqs)

            for ch_name in spectrum.ch_names:
                row_desc = (subject.name, key, ch_name)
                row_descs.append(row_desc)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = selected_name + '_all_subjects_all_channels_spectrum.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)

@threaded
def save_channel_averages(experiment, selected_name, channel_groups, log_transformed=False):
    """ Saves channel averages of a spectrum item to a csv file. """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        spectrum = subject.spectrum.get(selected_name)
        if not spectrum:
            continue

        ch_names = spectrum.ch_names
        freqs = spectrum.freqs
        info = spectrum.info

        for key, psd in spectrum.content.items():

            data_labels, averaged_data = average_to_channel_groups(
                psd, info, ch_names, channel_groups)

            if log_transformed:
                csv_data.extend(10 * np.log10(averaged_data.tolist()))
            else:
                csv_data.extend(averaged_data.tolist())

            column_names = format_floats(freqs)

            for ch_type, area in data_labels:
                row_desc = (subject.name, key, ch_type, area)
                row_descs.append(row_desc)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = selected_name + '_all_subjects_channel_averages_spectrum.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)




""" Contains a class for logic of montage dialog.
"""

import os
import pkg_resources
import logging

import mne

from copy import deepcopy

from PyQt5 import QtCore
from PyQt5 import QtWidgets

import meggie.utilities.filemanager as filemanager

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.actions.raw_montage.dialogs.montageDialogUi import Ui_montageDialog


class MontageDialog(QtWidgets.QDialog):
    """ Contains logic for montage dialog.
    """
    def __init__(self, parent, experiment, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_montageDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment
        self.handler = handler

        montage_dir = pkg_resources.resource_filename(
            'mne', os.path.join('channels', 'data', 'montages'))

        for fname in sorted(os.listdir(montage_dir)):
            self.ui.comboBoxSelectFromList.addItem(fname)
         
        self.current_montage_fname = None
        self.ui.radioButtonMontageFromList.setChecked(True)

        # find out if montage already set.

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

    def _experiment_getter(self):
        return self.experiment

    def on_pushButtonSelectFromFile_clicked(self, checked=None):
        if checked is None:
            return

        home = filemanager.homepath()

        fname = QtCore.QDir.toNativeSeparators(
            str(QtWidgets.QFileDialog.getOpenFileName(
                self, 'Open file', home, 
                "Montage-files (*.txt *.elc *.sfp);;"
                "All files (*.*)")[0]))

        self.current_montage_fname = fname
        if self.ui.radioButtonMontageFromFile.isChecked():
            self.ui.labelCurrentContent.setText(self.current_montage_fname)

    def on_radioButtonMontageFromFile_toggled(self, checked):
        if checked:
            self.ui.labelCurrentContent.setText(self.current_montage_fname)

    def on_comboBoxSelectFromList_currentIndexChanged(self, item):
        if self.ui.radioButtonMontageFromList.isChecked():
            self.ui.labelCurrentContent.setText(str(item))

    def on_radioButtonMontageFromList_toggled(self, checked):
        if checked:
            selection = self.ui.comboBoxSelectFromList.currentText()
            self.ui.labelCurrentContent.setText(selection)

    def _get_params(self):
        params = {}
        head_size = float(self.ui.doubleSpinBoxHeadSize.value())
        if self.ui.radioButtonMontageFromList.isChecked():
            selection = self.ui.comboBoxSelectFromList.currentText()
            selection = os.path.splitext(selection)[0]
            params['custom'] = False
            params['selection'] = selection
        else:
            params['custom'] = True
            params['selection'] = self.current_montage_fname

        params['head_size'] = head_size

        return params

    def accept(self):
        subject = self.experiment.active_subject

        params = self._get_params()

        try:
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):
        experiment = self.experiment
        selected_subject_names = self.batching_widget.selected_subjects

        params = self._get_params()

        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                except Exception as exc:
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        self.parent.initialize_ui()
        self.close()




""" Contains implementation for epochs plot
"""
import logging

import matplotlib.pyplot as plt
import numpy as np

from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class PlotEpochs(Action):
    """ Plots all epochs from all channels.
    """

    def run(self):

        try:
            selected_name = self.data['outputs']['epochs'][0]
        except IndexError as exc:
            return

        subject = self.experiment.active_subject
        try:
            self.handler(subject, {'name': selected_name})
        except Exception as exc:
            exc_messagebox(self.window, exc)


    @subject_action
    def handler(self, subject, params):
        epochs = subject.epochs.get(params['name'])
        epochs.content.plot()




"""Contains functions for tasks related to file system.
"""

import os
import shutil
import re
import datetime
import logging
import errno

import numpy as np
import mne


def open_raw(fname, preload=True, verbose='info'):
    """Reads a raw from file.

    Parameters
    ----------
    fname : str
        Path to the raw file.
    preload : bool
        Should the data be loaded as well or just the metadata.
    verbose : str
        Verbose level for the read_raw call.

    Returns
    -------
    mne.io.Raw
        The raw object read from the file.
    """
    try:
        if verbose == 'info' or verbose == 'debug':
            logging.getLogger('ui_logger').info('Reading ' + fname)
            raw = mne.io.read_raw(fname, preload=preload)
        else:
            raw = mne.io.read_raw(fname, preload=preload, verbose=verbose)
        return raw
    except Exception as exc:
        logging.getLogger('ui_logger').exception('')
        raise Exception('Could not read the raw file: ' + str(fname))

def save_raw(raw, path, overwrite=True):
    """Saves a raw to file(s).

    After witnessing several corruptions along the way, 
    this was made more atomic by saving the raw first to a tmp file 
    and then moving with shutil.

    Parameters
    ----------
    raw : mne.io.Raw
        The raw file to be saved.
    path : str
        The path where to save.
    overwrite : bool
        Whether to overwrite.
    """

    folder = os.path.dirname(path)
    bname = os.path.basename(path)

    exists = False
    if os.path.exists(path):
        if not overwrite:
            raise IOError('File already exists.')
        exists = True

    # be protective and save with other name first and move afterwards
    temp_path = os.path.join(folder, '_' + bname)
    raw.save(temp_path, overwrite=True)

    stem, ext = os.path.splitext(bname)
    ext_len = len(ext)

    # assumes filename ends with .fif
    pat_old = re.compile(bname[:-ext_len] + r'(-[0-9]+)?' + bname[-ext_len:])
    pat_new = re.compile('_' + bname[:-ext_len] + r'(-[0-9]+)?' + bname[-ext_len:])

    contents = os.listdir(folder)
    old_files = [fname_ for fname_ in contents if pat_old.match(fname_)]
    new_files = [fname_ for fname_ in contents if pat_new.match(fname_)]

    if len(old_files) != len(new_files):
        logger = logging.getLogger('ui_logger')
        if exists:
            logger.warning("Be warned, amount of parts has changed!")
        logger.debug("Old parts: ")
        for part in old_files:
            logger.debug(part)
        logger.debug("New parts: ")
        for part in new_files:
            logger.debug(part)

    moved_paths = []
    for file_ in new_files:
        tmp_path = os.path.join(folder, os.path.basename(file_))
        new_path = os.path.join(folder, os.path.basename(file_)[1:])
        shutil.move(tmp_path, new_path)
        moved_paths.append(new_path)

    for file_ in old_files:
        old_file_path = os.path.join(folder, os.path.basename(file_))
        if old_file_path not in moved_paths:
            logger.warning('Removing unused part: ' + str(old_file_path))
            os.remove(old_file_path)

    raw._filenames[0] = path


def ensure_folders(paths):
    """Ensures that paths in specified in the paths param exist.

    Parameters
    ----------
    paths : list
        List of folder paths.
    """
    for path in paths:
        if not os.path.exists(path):
            os.makedirs(path)


def create_timestamped_folder(experiment):
    """Creates folder with a timestamp inside the output folder in the
    experiment folder.

    Parameters
    ----------
    experiment : meggie.experiment.Experiment
        The experiment where to create the folder.

    Returns
    -------
    str
        The path to the folder.
    """
    current_time_str = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    path = os.path.join(experiment.path, 'output')
    timestamped_folder = os.path.join(path, current_time_str)

    try:
        os.makedirs(timestamped_folder)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    return timestamped_folder


def save_csv(path, data, column_names, row_descs):
    """ Saves tabular data to csv.

    Parameters
    ----------
    path : str
        Where to save.
    data : a numpy array of shape (n_rows, n_cols)
        Data to save.
    column_names : list
        List of column names.
    row_descs : list
        List of row descriptions that can be tuples like
        ('EEG', 'Left-frontal'), which are then put to the 
        csv as multiple columns.
    """
    # gather all the data to list of rows
    all_data = []

    if type(data) == np.ndarray:
        data = data.tolist()

    # freqs data, assume same lengths
    all_data.append(['']*len(row_descs[0]) + column_names)

    for idx in range(len(data)):
        row = list(row_descs[idx]) + data[idx]
        all_data.append(row)

    # save to file
    all_data = np.array(all_data)
    np.savetxt(path, all_data, fmt='%s', delimiter=', ')


def load_csv(path):
    """Loads tabular data from csv.

    Parameters
    ----------
    path : str
        Path to the csv file.

    Returns
    -------
    list
        Column names.
    list
        Row descriptions.
    np.array
        The data.
    """
    all_data = np.loadtxt(path, dtype=np.str, delimiter=', ')

    data = []
    column_names = []
    row_descs = []

    first_data_idx = np.min(np.where(all_data[0] != '')[0])

    column_names = all_data[0, first_data_idx:].tolist()
    row_descs = [tuple(elems) for elems in all_data[1:, :first_data_idx]]
    data = all_data[1:, first_data_idx:].astype(np.float)

    return column_names, row_descs, data


def tail(f, lines=1, _buffer=4098):
    """Tail a file and get `lines` lines from the end,

    See https://stackoverflow.com/a/13790289

    Parameters
    ----------
    f : file descriptor
        The file descriptor already opened.
    lines : int
        How many lines to read from the end.
    _buffer : int
        What buffer size to use.

    Returns
    -------
    list
        Lines from the end.
    """
    # place holder for the lines found
    lines_found = []

    # block counter will be multiplied by buffer
    # to get the block size from the end
    block_counter = -1

    # loop until we find X lines
    while len(lines_found) < lines:
        try:
            f.seek(block_counter * _buffer, os.SEEK_END)
        except IOError:  # either file is too small, or too many lines requested
            f.seek(0)
            lines_found = f.readlines()
            break

        lines_found = f.readlines()

        block_counter -= 1

    return lines_found[-lines:]


def homepath():
    """Tries to find correct path for home folder.

    Returns
    -------
    str
        Path to home directory.

    """
    from os.path import expanduser
    home = expanduser("~")

    if not home:
        return '.'

    return home




""" Contains plot tse action handling.
"""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.names import next_available_name
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.utilities.dialogs.TFROutputOptionsMain import TFROutputOptions

from meggie.actions.tfr_plot_tse.controller.tfr import plot_tse_averages
from meggie.actions.tfr_plot_tse.controller.tfr import plot_tse_topo


class PlotTSE(Action):

    def run(self):
        """ Plots a TSE from TFR object.
        """
        try:
            selected_name = self.data['outputs']['tfr'][0]
        except IndexError as exc:
            return

        def option_handler(params):
            params['channel_groups'] = self.experiment.channel_groups
            params['name'] = selected_name

            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = TFROutputOptions(self.window, self.experiment,
                                  selected_name, handler=option_handler,
                                  ask_condition=False)
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """
        """
        info = subject.tfr[params['name']].info
        if params['output_option'] == 'all_channels':
            chs = list(get_channels_by_type(info).keys())
            if 'eeg' in chs:
                plot_tse_topo(subject, params['name'],
                              params['blmode'], params['blstart'], params['blend'],
                              params['tmin'], params['tmax'], params['fmin'], params['fmax'],
                              'eeg')
            if 'grad' in chs or 'mag' in chs:
                plot_tse_topo(subject, params['name'],
                              params['blmode'], params['blstart'], params['blend'],
                              params['tmin'], params['tmax'], params['fmin'], params['fmax'],
                              'meg')
        else:
            plot_tse_averages(subject, params['name'],
                              params['blmode'], params['blstart'], params['blend'],
                              params['tmin'], params['tmax'], params['fmin'], params['fmax'],
                              params['channel_groups'])




""" Contains create tfr action handling.
"""

from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.tfr_create.dialogs.TFRDialogMain import TFRDialog
from meggie.actions.tfr_create.controller.tfr import create_tfr


class CreateTFR(Action):
    """ Shows a TFR parameter dialog and then
    creates TFR """

    def run(self):

        selected_names = self.data['inputs']['epochs']

        if not selected_names:
            return

        if len(selected_names) == 1:
            stem = selected_names[0]
        else:
            stem = 'TFR'

        default_name = next_available_name(
            self.experiment.active_subject.tfr.keys(), stem)

        subject = self.experiment.active_subject

        dialog = TFRDialog(self.experiment, self.window, 
                            selected_names, default_name, self.handler)
        dialog.show()


    @subject_action
    def handler(self, subject, params):
        """
        """
        tfr_name = params['name']
        epochs_names = params['conditions']
        freqs = params['freqs']
        decim = params['decim']
        n_cycles = params['n_cycles']
        subtract_evoked = params['subtract_evoked']
        
        create_tfr(subject, tfr_name, epochs_names, freqs, decim, 
                   n_cycles, subtract_evoked, do_meanwhile=self.window.update_ui)




""" Contains implementation for tfr info
"""
from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float
from meggie.utilities.formats import format_floats


class Info(InfoAction):
    """ Fills up tfr info box """

    def run(self):
        try:
            selected_name = self.data['outputs']['tfr'][0]

            tfr = self.experiment.active_subject.tfr[selected_name]
            params = tfr.params

            message = ""
            if 'decim' in params:
                message += 'Decimated by factor: {0}\n'.format(params['decim'])
            if 'evoked_subtracted' in params:
                message += 'Evoked subtracted: {0}\n'.format(params['evoked_subtracted'])
            if 'conditions' in params:
                message += 'Conditions: ' + ', '.join(params['conditions']) + '\n'
            if 'n_cycles' in params:
                message += 'Cycles: ' + ', '.join(format_floats(params['n_cycles'])) + '\n'

            if hasattr(tfr, 'times'):
                message += 'Times: {0}s - {1}s\n'.format(
                    format_float(tfr.times[0]), format_float(tfr.times[-1]))

            if hasattr(tfr, 'freqs'):
                message += 'Freqs: {0} - {1}\n'.format(
                    format_float(tfr.freqs[0]), format_float(tfr.freqs[-1]))

            if 'groups' in params:
                for key, names in params['groups'].items():
                    message += '\nGroup ' + str(key) + ': \n'
                    for name in names:
                        message += name + '\n'

        except Exception as exc:
            message = ""

        return message




""" Contains ICA action handling.
"""
import logging 

import numpy as np
import matplotlib.pyplot as plt

from meggie.actions.raw_ica.dialogs.icaDialogMain import ICADialog

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class ICA(Action):
    """ Shows a dialog which can be used to compute ICA decompositions
    and selecting components for removal.
    """
    random_state = 10

    def run(self):
        ica_dialog = ICADialog(self.window, self.experiment, 
                               self.random_state, on_apply=self.handler)
        ica_dialog.show()

    @subject_action
    def handler(self, subject, params):
        """
        """




""" Contains a class for logic of pipeline dialog.
"""

import os
import json
import pkg_resources

from PyQt5 import QtWidgets

from meggie.mainwindow.dynamic import find_all_sources
from meggie.mainwindow.dynamic import find_all_package_specs

from meggie.mainwindow.dialogs.pipelineDialogUi import Ui_pipelineDialog

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox


class PipelineDialog(QtWidgets.QDialog):
    """ Contains logic for pipeline dialog.
    """

    def __init__(self, parent, prefs):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_pipelineDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = parent.experiment

        # Read selected pipeline from the experiment
        selected_pipeline = self.experiment.selected_pipeline

        self.active_plugins = prefs.active_plugins

        # read all pipeline ids and names to a list
        pipelines = []
        package_specs = find_all_package_specs()

        for source, package_spec in package_specs.items():
            if source == 'meggie' or source in self.active_plugins:

                if 'pipelines' in package_spec:
                    for pipeline in package_spec['pipelines']:
                        try:
                            id_ = pipeline['id']
                        except Exception as exc:
                            raise Exception('Every pipeline should have id.')

                        name = pipeline.get('name', '')
                        pipelines.append((id_, name))

        pipelines.append(('classic', 'Include everything'))

        self.pipelines = pipelines

        # create buttons for pipelines
        self.pipeline_buttons = []
        for idx, (pipeline_id, pipeline_name) in enumerate(self.pipelines):
            radio_button = QtWidgets.QRadioButton(self.ui.groupBoxPipeline)
            radio_button.setText(pipeline_name)

            self.ui.gridLayoutPipeline.addWidget(
                radio_button, idx + 1, 0, 1, 1)

            self.pipeline_buttons.append(radio_button)

            if selected_pipeline == pipeline_id:
                radio_button.setChecked(True)

        if len(self.pipeline_buttons) == 1:
            self.pipeline_buttons[0].setEnabled(False)
            self.pipeline_buttons[0].setChecked(True)

    def accept(self):

        selected_pipeline = ""
        for button_idx, radio_button in enumerate(self.pipeline_buttons):
            if radio_button.isChecked():
                selected_pipeline = self.pipelines[button_idx][0]
                break

        # store selected pipeline to the experiment
        if not selected_pipeline:
            return

        try:
            self.experiment.selected_pipeline = selected_pipeline
            self.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.reconstruct_tabs()
        self.parent.initialize_ui()

        self.close()



""" Contains a class for logic of preferences dialog.
"""

import os
import json
import logging
import pkg_resources

from PyQt5 import QtCore
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal

from meggie.mainwindow.dialogs.preferencesDialogUi import Ui_DialogPreferences

from meggie.mainwindow.dialogs.activePluginsDialogMain import ActivePluginsDialog

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox


class PreferencesDialog(QtWidgets.QDialog):
    """ Contains logic for preferences dialog.
    """

    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_DialogPreferences()
        self.ui.setupUi(self)

        self.parent = parent
        self.prefs = self.parent.prefs

        # Prefill previous values to UI and attributes from config file.
        self.ui.LineEditFilePath.setText(self.prefs.workspace)

        if self.prefs.auto_load_last_open_experiment:
            self.ui.checkBoxAutomaticOpenPreviousExperiment.setChecked(True)

        self.active_plugins = self.prefs.active_plugins

    def on_ButtonBrowseWorkingDir_clicked(self, checked=None):
        if checked is None:
            return

        workspace = QtCore.QDir.toNativeSeparators(
            str(QtWidgets.QFileDialog.getExistingDirectory(
                self, "Select a workspace directory")))
        self.ui.LineEditFilePath.setText(workspace)

    def on_pushButtonPlugins_clicked(self, checked=None):
        if checked is None:
            return

        def handler(selected_plugins):
            self.active_plugins = selected_plugins

        dialog = ActivePluginsDialog(self.active_plugins, self, handler)
        dialog.show()

    def accept(self):
        workspace = self.ui.LineEditFilePath.text()
        if not os.path.isdir(workspace):
            message = 'Workspace must be set to proper path.'
            messagebox(self.parent, message)
            return
        self.prefs.workspace = workspace

        if self.ui.checkBoxAutomaticOpenPreviousExperiment.isChecked():
            autoLoadLastOpenExp = True
        else:
            autoLoadLastOpenExp = False
        self.prefs.auto_load_last_open_experiment = autoLoadLastOpenExp  # noqa

        self.prefs.active_plugins = self.active_plugins

        try:
            self.prefs.write_preferences_to_disk()
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        # Plugins can add new actions to existing pipelines.
        self.parent.reconstruct_tabs()
        self.parent.initialize_ui()

        self.close()



""" Contains tfr group average action handling.
"""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.tfr_group_average.controller.tfr import group_average_tfr

from meggie.utilities.dialogs.groupSelectionDialogMain import GroupSelectionDialog


class GroupAverage(Action):
    """ Creates group average items
    """

    def run(self):
        """
        """
        try:
            selected_name = self.data['outputs']['tfr'][0]
        except IndexError as exc:
            return

        name = next_available_name(
            self.experiment.active_subject.tfr.keys(),
            'group_' + selected_name)

        def group_handler(groups):
            params = {'based_on': selected_name, 
                      'name': name,
                      'groups': groups}
            try:
                self.handler(self.experiment.active_subject, params)
                self.experiment.save_experiment_settings()
                self.window.initialize_ui()
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = GroupSelectionDialog(self.experiment, self.window, handler=group_handler)
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """
        """
        group_average_tfr(self.experiment, 
                             params['based_on'],
                             params['groups'],
                             params['name'],
                             do_meanwhile=self.window.update_ui)





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'powerSpectrumAddAdvancedDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PowerSpectrumAddAdvancedDialog(object):
    def setupUi(self, PowerSpectrumAddAdvancedDialog):
        PowerSpectrumAddAdvancedDialog.setObjectName("PowerSpectrumAddAdvancedDialog")
        PowerSpectrumAddAdvancedDialog.resize(356, 546)
        self.gridLayout_2 = QtWidgets.QGridLayout(PowerSpectrumAddAdvancedDialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBoxGeneral = QtWidgets.QGroupBox(PowerSpectrumAddAdvancedDialog)
        self.groupBoxGeneral.setObjectName("groupBoxGeneral")
        self.formLayout = QtWidgets.QFormLayout(self.groupBoxGeneral)
        self.formLayout.setObjectName("formLayout")
        self.labelGroup = QtWidgets.QLabel(self.groupBoxGeneral)
        self.labelGroup.setObjectName("labelGroup")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelGroup)
        self.comboBoxAvgGroup = QtWidgets.QComboBox(self.groupBoxGeneral)
        self.comboBoxAvgGroup.setObjectName("comboBoxAvgGroup")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.comboBoxAvgGroup.addItem("")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.comboBoxAvgGroup)
        self.gridLayout_2.addWidget(self.groupBoxGeneral, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pushButtonCancel = QtWidgets.QPushButton(PowerSpectrumAddAdvancedDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonAccept = QtWidgets.QPushButton(PowerSpectrumAddAdvancedDialog)
        self.pushButtonAccept.setObjectName("pushButtonAccept")
        self.horizontalLayout.addWidget(self.pushButtonAccept)
        self.gridLayout_2.addLayout(self.horizontalLayout, 4, 0, 1, 1)
        self.groupBoxStartingPoints = QtWidgets.QGroupBox(PowerSpectrumAddAdvancedDialog)
        self.groupBoxStartingPoints.setObjectName("groupBoxStartingPoints")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBoxStartingPoints)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.labelStartId = QtWidgets.QLabel(self.groupBoxStartingPoints)
        self.labelStartId.setObjectName("labelStartId")
        self.gridLayout_3.addWidget(self.labelStartId, 1, 0, 1, 1)
        self.labelStartOffset = QtWidgets.QLabel(self.groupBoxStartingPoints)
        self.labelStartOffset.setObjectName("labelStartOffset")
        self.gridLayout_3.addWidget(self.labelStartOffset, 5, 0, 1, 1)
        self.labelStartMask = QtWidgets.QLabel(self.groupBoxStartingPoints)
        self.labelStartMask.setObjectName("labelStartMask")
        self.gridLayout_3.addWidget(self.labelStartMask, 2, 0, 1, 1)
        self.pushButtonStartEdit = QtWidgets.QPushButton(self.groupBoxStartingPoints)
        self.pushButtonStartEdit.setEnabled(False)
        self.pushButtonStartEdit.setObjectName("pushButtonStartEdit")
        self.gridLayout_3.addWidget(self.pushButtonStartEdit, 1, 3, 1, 1)
        self.radioButtonStartUseEvents = QtWidgets.QRadioButton(self.groupBoxStartingPoints)
        self.radioButtonStartUseEvents.setObjectName("radioButtonStartUseEvents")
        self.gridLayout_3.addWidget(self.radioButtonStartUseEvents, 0, 0, 1, 4)
        self.radioButtonStartUseStart = QtWidgets.QRadioButton(self.groupBoxStartingPoints)
        self.radioButtonStartUseStart.setChecked(True)
        self.radioButtonStartUseStart.setObjectName("radioButtonStartUseStart")
        self.gridLayout_3.addWidget(self.radioButtonStartUseStart, 3, 0, 1, 4)
        self.doubleSpinBoxStartOffset = QtWidgets.QDoubleSpinBox(self.groupBoxStartingPoints)
        self.doubleSpinBoxStartOffset.setMinimum(-100000000.0)
        self.doubleSpinBoxStartOffset.setMaximum(100000000.0)
        self.doubleSpinBoxStartOffset.setProperty("value", 5.0)
        self.doubleSpinBoxStartOffset.setObjectName("doubleSpinBoxStartOffset")
        self.gridLayout_3.addWidget(self.doubleSpinBoxStartOffset, 5, 2, 1, 1)
        self.spinBoxStartId = QtWidgets.QSpinBox(self.groupBoxStartingPoints)
        self.spinBoxStartId.setEnabled(False)
        self.spinBoxStartId.setMaximum(10000000)
        self.spinBoxStartId.setProperty("value", 1)
        self.spinBoxStartId.setObjectName("spinBoxStartId")
        self.gridLayout_3.addWidget(self.spinBoxStartId, 1, 2, 1, 1)
        self.spinBoxStartMask = QtWidgets.QSpinBox(self.groupBoxStartingPoints)
        self.spinBoxStartMask.setEnabled(False)
        self.spinBoxStartMask.setMaximum(1000000000)
        self.spinBoxStartMask.setObjectName("spinBoxStartMask")
        self.gridLayout_3.addWidget(self.spinBoxStartMask, 2, 2, 1, 1)
        self.radioButtonStartUseEnd = QtWidgets.QRadioButton(self.groupBoxStartingPoints)
        self.radioButtonStartUseEnd.setObjectName("radioButtonStartUseEnd")
        self.gridLayout_3.addWidget(self.radioButtonStartUseEnd, 4, 0, 1, 4)
        self.gridLayout_2.addWidget(self.groupBoxStartingPoints, 1, 0, 1, 1)
        self.groupBoxEndingPoints = QtWidgets.QGroupBox(PowerSpectrumAddAdvancedDialog)
        self.groupBoxEndingPoints.setObjectName("groupBoxEndingPoints")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBoxEndingPoints)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.doubleSpinBoxEndOffset = QtWidgets.QDoubleSpinBox(self.groupBoxEndingPoints)
        self.doubleSpinBoxEndOffset.setMinimum(-10000000.0)
        self.doubleSpinBoxEndOffset.setMaximum(10000000.99)
        self.doubleSpinBoxEndOffset.setProperty("value", -5.0)
        self.doubleSpinBoxEndOffset.setObjectName("doubleSpinBoxEndOffset")
        self.gridLayout_4.addWidget(self.doubleSpinBoxEndOffset, 6, 2, 1, 1)
        self.labelEndMask = QtWidgets.QLabel(self.groupBoxEndingPoints)
        self.labelEndMask.setObjectName("labelEndMask")
        self.gridLayout_4.addWidget(self.labelEndMask, 2, 0, 1, 1)
        self.labelEndId = QtWidgets.QLabel(self.groupBoxEndingPoints)
        self.labelEndId.setObjectName("labelEndId")
        self.gridLayout_4.addWidget(self.labelEndId, 1, 0, 1, 1)
        self.labelEndOffset = QtWidgets.QLabel(self.groupBoxEndingPoints)
        self.labelEndOffset.setObjectName("labelEndOffset")
        self.gridLayout_4.addWidget(self.labelEndOffset, 6, 0, 1, 1)
        self.pushButtonEndEdit = QtWidgets.QPushButton(self.groupBoxEndingPoints)
        self.pushButtonEndEdit.setEnabled(False)
        self.pushButtonEndEdit.setObjectName("pushButtonEndEdit")
        self.gridLayout_4.addWidget(self.pushButtonEndEdit, 1, 3, 1, 1)
        self.radioButtonEndUseEvents = QtWidgets.QRadioButton(self.groupBoxEndingPoints)
        self.radioButtonEndUseEvents.setObjectName("radioButtonEndUseEvents")
        self.gridLayout_4.addWidget(self.radioButtonEndUseEvents, 0, 0, 1, 4)
        self.radioButtonEndUseEnd = QtWidgets.QRadioButton(self.groupBoxEndingPoints)
        self.radioButtonEndUseEnd.setChecked(True)
        self.radioButtonEndUseEnd.setObjectName("radioButtonEndUseEnd")
        self.gridLayout_4.addWidget(self.radioButtonEndUseEnd, 4, 0, 1, 4)
        self.spinBoxEndId = QtWidgets.QSpinBox(self.groupBoxEndingPoints)
        self.spinBoxEndId.setEnabled(False)
        self.spinBoxEndId.setMaximum(10000000)
        self.spinBoxEndId.setProperty("value", 2)
        self.spinBoxEndId.setObjectName("spinBoxEndId")
        self.gridLayout_4.addWidget(self.spinBoxEndId, 1, 2, 1, 1)
        self.spinBoxEndMask = QtWidgets.QSpinBox(self.groupBoxEndingPoints)
        self.spinBoxEndMask.setEnabled(False)
        self.spinBoxEndMask.setMaximum(1000000000)
        self.spinBoxEndMask.setObjectName("spinBoxEndMask")
        self.gridLayout_4.addWidget(self.spinBoxEndMask, 2, 2, 1, 1)
        self.radioButtonEndUseStart = QtWidgets.QRadioButton(self.groupBoxEndingPoints)
        self.radioButtonEndUseStart.setObjectName("radioButtonEndUseStart")
        self.gridLayout_4.addWidget(self.radioButtonEndUseStart, 3, 0, 1, 4)
        self.gridLayout_2.addWidget(self.groupBoxEndingPoints, 2, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem1, 3, 0, 1, 1)

        self.retranslateUi(PowerSpectrumAddAdvancedDialog)
        self.pushButtonCancel.clicked.connect(PowerSpectrumAddAdvancedDialog.reject)
        self.pushButtonAccept.clicked.connect(PowerSpectrumAddAdvancedDialog.accept)
        self.radioButtonEndUseEvents.toggled['bool'].connect(self.spinBoxEndId.setEnabled)
        self.radioButtonEndUseEvents.toggled['bool'].connect(self.spinBoxEndMask.setEnabled)
        self.radioButtonEndUseEvents.toggled['bool'].connect(self.pushButtonEndEdit.setEnabled)
        self.radioButtonStartUseEvents.toggled['bool'].connect(self.spinBoxStartId.setEnabled)
        self.radioButtonStartUseEvents.toggled['bool'].connect(self.spinBoxStartMask.setEnabled)
        self.radioButtonStartUseEvents.toggled['bool'].connect(self.pushButtonStartEdit.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(PowerSpectrumAddAdvancedDialog)

    def retranslateUi(self, PowerSpectrumAddAdvancedDialog):
        _translate = QtCore.QCoreApplication.translate
        PowerSpectrumAddAdvancedDialog.setWindowTitle(_translate("PowerSpectrumAddAdvancedDialog", "Meggie - Add advanced"))
        self.groupBoxGeneral.setTitle(_translate("PowerSpectrumAddAdvancedDialog", "General:"))
        self.labelGroup.setText(_translate("PowerSpectrumAddAdvancedDialog", "Average group:"))
        self.comboBoxAvgGroup.setItemText(0, _translate("PowerSpectrumAddAdvancedDialog", "1"))
        self.comboBoxAvgGroup.setItemText(1, _translate("PowerSpectrumAddAdvancedDialog", "2"))
        self.comboBoxAvgGroup.setItemText(2, _translate("PowerSpectrumAddAdvancedDialog", "3"))
        self.comboBoxAvgGroup.setItemText(3, _translate("PowerSpectrumAddAdvancedDialog", "4"))
        self.comboBoxAvgGroup.setItemText(4, _translate("PowerSpectrumAddAdvancedDialog", "5"))
        self.comboBoxAvgGroup.setItemText(5, _translate("PowerSpectrumAddAdvancedDialog", "6"))
        self.comboBoxAvgGroup.setItemText(6, _translate("PowerSpectrumAddAdvancedDialog", "7"))
        self.comboBoxAvgGroup.setItemText(7, _translate("PowerSpectrumAddAdvancedDialog", "8"))
        self.pushButtonCancel.setText(_translate("PowerSpectrumAddAdvancedDialog", "Cancel"))
        self.pushButtonAccept.setText(_translate("PowerSpectrumAddAdvancedDialog", "Add"))
        self.groupBoxStartingPoints.setTitle(_translate("PowerSpectrumAddAdvancedDialog", "Starting points:"))
        self.labelStartId.setText(_translate("PowerSpectrumAddAdvancedDialog", "Event id:"))
        self.labelStartOffset.setText(_translate("PowerSpectrumAddAdvancedDialog", "Offset:"))
        self.labelStartMask.setText(_translate("PowerSpectrumAddAdvancedDialog", "Mask:"))
        self.pushButtonStartEdit.setText(_translate("PowerSpectrumAddAdvancedDialog", "Edit..."))
        self.radioButtonStartUseEvents.setText(_translate("PowerSpectrumAddAdvancedDialog", "Get starting points from events:"))
        self.radioButtonStartUseStart.setText(_translate("PowerSpectrumAddAdvancedDialog", "Use start of recording"))
        self.doubleSpinBoxStartOffset.setSuffix(_translate("PowerSpectrumAddAdvancedDialog", "s"))
        self.radioButtonStartUseEnd.setText(_translate("PowerSpectrumAddAdvancedDialog", "Use end of recording"))
        self.groupBoxEndingPoints.setTitle(_translate("PowerSpectrumAddAdvancedDialog", "Ending points:"))
        self.doubleSpinBoxEndOffset.setSuffix(_translate("PowerSpectrumAddAdvancedDialog", "s"))
        self.labelEndMask.setText(_translate("PowerSpectrumAddAdvancedDialog", "Mask:"))
        self.labelEndId.setText(_translate("PowerSpectrumAddAdvancedDialog", "Event id:"))
        self.labelEndOffset.setText(_translate("PowerSpectrumAddAdvancedDialog", "Offset:"))
        self.pushButtonEndEdit.setText(_translate("PowerSpectrumAddAdvancedDialog", "Edit..."))
        self.radioButtonEndUseEvents.setText(_translate("PowerSpectrumAddAdvancedDialog", "Get starting points from events:"))
        self.radioButtonEndUseEnd.setText(_translate("PowerSpectrumAddAdvancedDialog", "Use end of recording"))
        self.radioButtonEndUseStart.setText(_translate("PowerSpectrumAddAdvancedDialog", "Use start of recording"))




""" Contains controlling logic for the evoked implementation.
"""

import logging
import os

import mne
import numpy as np

import meggie.utilities.filemanager as filemanager

from meggie.utilities.formats import format_floats
from meggie.utilities.channels import average_to_channel_groups

from meggie.utilities.threading import threaded


def _create_averages(mne_evoked, channel_groups):

    mne_evoked = mne_evoked.copy().drop_channels(mne_evoked.info['bads'])

    data_labels, averaged_data = average_to_channel_groups(
        mne_evoked.data, mne_evoked.info, 
        mne_evoked.info['ch_names'], channel_groups)

    return data_labels, averaged_data


@threaded
def save_all_channels(experiment, selected_name):
    """ Saves all channels of evoked item to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        evoked = subject.evoked.get(selected_name)
        if not evoked:
            continue

        for key, mne_evoked in evoked.content.items():
            column_names = format_floats(mne_evoked.times)

            for ch_idx, ch_name in enumerate(mne_evoked.info['ch_names']):
                if ch_name in mne_evoked.info['bads']:
                    continue
                csv_data.append(mne_evoked.data[ch_idx].tolist())

                row_desc = (subject.name, key, ch_name)
                row_descs.append(row_desc)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = selected_name + '_all_subjects_all_channels_evoked.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)

@threaded
def save_channel_averages(experiment, selected_name, channel_groups):
    """ Saves channel averages to a csv file.
    """
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        evoked = subject.evoked.get(selected_name)
        if not evoked:
            continue

        for key, mne_evoked in evoked.content.items():

            data_labels, averaged_data = _create_averages(
                mne_evoked, channel_groups)

            csv_data.extend(averaged_data.tolist())
            column_names = format_floats(mne_evoked.times)

            for ch_type, area in data_labels:
                row_desc = (subject.name, key, ch_type, area)
                row_descs.append(row_desc)

    folder = filemanager.create_timestamped_folder(experiment)
    fname = selected_name + '_all_subjects_channel_averages_evoked.csv'
    path = os.path.join(folder, fname)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger('ui_logger').info('Saved the csv file to ' + path)




""" Contains implementation for epochs plot
"""
import logging

import matplotlib.pyplot as plt
import numpy as np

from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class PlotEpochsImage(Action):
    """ Opens MNE's image plot.
    """

    def run(self):

        try:
            selected_name = self.data['outputs']['epochs'][0]
        except IndexError as exc:
            return

        subject = self.experiment.active_subject
        try:
            self.handler(subject, {'name': selected_name})
        except Exception as exc:
            exc_messagebox(self.window, exc)


    @subject_action
    def handler(self, subject, params):
        epochs = subject.epochs.get(params['name'])
        figs = epochs.content.plot_image()
        for fig in figs:
            ch_type = fig.canvas.get_window_title()
            title = '{0}_{1}'.format(params['name'], ch_type)
            fig.canvas.set_window_title(title)




""" Contains a class for logic of the batching widget.
"""

from PyQt5 import QtCore
from PyQt5 import QtWidgets

from meggie.utilities.widgets.batchingWidgetUi import Ui_BatchingWidget
from meggie.utilities.messaging import messagebox


class BatchingWidget(QtWidgets.QWidget):
    """ Contains logic for batching widget.
    """

    def __init__(self, experiment_getter, parent, geometry,
                 container=None,
                 pushButtonApply=None,
                 pushButtonBatch=None,
                 hideHook=None):
        super(BatchingWidget, self).__init__(container)
        self.ui = Ui_BatchingWidget()
        self.ui.setupUi(self)
        self.parent = parent

        if not pushButtonApply:
            pushButtonApply = self.parent.ui.pushButtonApply
        if not pushButtonBatch:
            pushButtonBatch = self.parent.ui.pushButtonBatch

        self.experiment_getter = experiment_getter
        self.experiment = None
        self.hideHook = hideHook

        self.pushButtonBatch = pushButtonBatch
        self.pushButtonApply = pushButtonApply

        self.pushButtonBatch.setEnabled(False)
        self.pushButtonApply.setEnabled(True)

        self.ui.functionalityWidget.hide()
        self.setGeometry(geometry)
        self.adjustSize()

        self.failed_subjects = []

    def _update(self, enabled):

        if not self.experiment:
            return

        if enabled:
            self.ui.functionalityWidget.show()
            self.adjustSize()
            self.pushButtonApply.setEnabled(False)
            self.pushButtonBatch.setEnabled(True)

            subject_names = sorted(self.experiment.subjects.keys())

            for name in subject_names:
                item = QtWidgets.QListWidgetItem(name)
                item.setCheckState(QtCore.Qt.Unchecked)
                item.setFlags(QtCore.Qt.ItemIsEnabled)
                self.ui.listWidgetSubjects.addItem(item)

        else:
            self.ui.listWidgetSubjects.clear()
            self.ui.functionalityWidget.hide()
            self.adjustSize()
            self.pushButtonApply.setEnabled(True)
            self.pushButtonBatch.setEnabled(False)

            if self.hideHook:
                self.hideHook()

    def on_listWidgetSubjects_itemClicked(self, item):
        if not item:
            return

        if item.checkState() != QtCore.Qt.Checked:
            item.setCheckState(QtCore.Qt.Checked)
        else:
            item.setCheckState(QtCore.Qt.Unchecked)

    def showWidget(self, enabled):
        self.experiment = self.experiment_getter()

        if self.experiment:
            self._update(enabled)

    def on_pushButtonApplyAll_clicked(self, checked=None):
        """
        """
        if checked is None:
            return

        for i in range(self.ui.listWidgetSubjects.count()):
            item = self.ui.listWidgetSubjects.item(i)
            item.setCheckState(QtCore.Qt.Checked)

    @property
    def selected_subjects(self):
        """ Returns the subjects that are selected in the UI.
        """
        subject_names = []
        for i in range(self.ui.listWidgetSubjects.count()):
            item = self.ui.listWidgetSubjects.item(i)
            if item.checkState() == QtCore.Qt.Checked:
                subject_names.append(item.text())

        return subject_names

    def cleanup(self):
        """ Cleans the batching widget after use.
        """
        if len(self.failed_subjects) > 0:
            rows = []
            rows.append('Failed calculation for subjects:')

            for subject, message in self.failed_subjects:
                rows.append(subject.name + ' (' + message + ')')

            parent = self.parent.parent

            messagebox(parent, '\n'.join(rows))

        self.failed_subjects = []
        self.ui.checkBoxBatch.setChecked(False)
        self.ui.functionalityWidget.hide()

